#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from copy import deepcopy

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import NoteType
from gramps.gen.db import DbTxn

from .displaytabs import NoteTab, AddrEmbedList, WebEmbedList, SourceBackRefList
from ..widgets import MonitoredEntry, PrivacyButton, MonitoredDataType
from .editreference import RefTab, EditReference
from ..glade import Glade


# -------------------------------------------------------------------------
#
# EditRepoRef class
#
# -------------------------------------------------------------------------
class EditRepoRef(EditReference):
    def __init__(self, state, uistate, track, source, source_ref, update):
        EditReference.__init__(self, state, uistate, track, source, source_ref, update)
        self.original = deepcopy(source.serialize())

    def _local_init(self):
        self.top = Glade()
        self.set_window(
            self.top.toplevel,
            self.top.get_object("repo_title"),
            _("Repository Reference Editor"),
        )
        self.setup_configs("interface.repo-ref", 600, 450)

        self.define_warn_box(self.top.get_object("warn_box"))
        self.define_expander(self.top.get_object("src_expander"))

        tblref = self.top.get_object("table70")
        notebook = self.top.get_object("notebook_ref")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.reftab = RefTab(
            self.dbstate, self.uistate, self.track, _("General"), tblref
        )
        tblref = self.top.get_object("table69")
        notebook = self.top.get_object("notebook_src")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.primtab = RefTab(
            self.dbstate, self.uistate, self.track, _("_General"), tblref
        )
        self.track_ref_for_deletion("primtab")

    def _connect_signals(self):
        self.define_ok_button(self.top.get_object("ok"), self.ok_clicked)
        self.define_cancel_button(self.top.get_object("cancel"))

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("repository-rebuild", self.close)
        self._add_db_signal("repository-delete", self.check_for_close)

    def _setup_fields(self):
        self.callno = MonitoredEntry(
            self.top.get_object("call_number"),
            self.source_ref.set_call_number,
            self.source_ref.get_call_number,
            self.db.readonly,
        )

        self.gid = MonitoredEntry(
            self.top.get_object("gid"),
            self.source.set_gramps_id,
            self.source.get_gramps_id,
            self.db.readonly,
        )

        self.privacy = PrivacyButton(
            self.top.get_object("private"), self.source, self.db.readonly
        )

        self.ref_privacy = PrivacyButton(
            self.top.get_object("private_ref"), self.source_ref, self.db.readonly
        )

        self.title = MonitoredEntry(
            self.top.get_object("repo_name"),
            self.source.set_name,
            self.source.get_name,
            self.db.readonly,
        )

        self.type_selector = MonitoredDataType(
            self.top.get_object("media_type"),
            self.source_ref.set_media_type,
            self.source_ref.get_media_type,
            self.db.readonly,
            self.db.get_source_media_types(),
        )

        self.media_type_selector = MonitoredDataType(
            self.top.get_object("repo_type"),
            self.source.set_type,
            self.source.get_type,
            self.db.readonly,
            self.db.get_repository_types(),
        )

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.
        """

        notebook_src = self.top.get_object("notebook_src")
        notebook_ref = self.top.get_object("notebook_ref")

        self._add_tab(notebook_src, self.primtab)
        self._add_tab(notebook_ref, self.reftab)

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_note_list(),
            "reporef_editor_shared_notes",
            notetype=NoteType.REPO,
        )
        self._add_tab(notebook_src, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.comment_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source_ref.get_note_list(),
            "reporef_editor_ref_notes",
            notetype=NoteType.REPOREF,
        )
        self._add_tab(notebook_ref, self.comment_tab)
        self.track_ref_for_deletion("comment_tab")

        self.address_tab = AddrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_address_list(),
            "reporef_editor_shared_address",
        )
        self._add_tab(notebook_src, self.address_tab)
        self.track_ref_for_deletion("address_tab")

        self.web_list = WebEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_url_list(),
            "reporef_editor_shared_internet",
        )
        self._add_tab(notebook_src, self.web_list)
        self.track_ref_for_deletion("web_list")

        self.backref_tab = SourceBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.source.handle),
            "reporef_editor_shared_references",
            self.enable_warnbox,
        )
        self._add_tab(notebook_src, self.backref_tab)
        self.track_ref_for_deletion("backref_tab")

        self._setup_notebook_tabs(notebook_src)
        self._setup_notebook_tabs(notebook_ref)

    def build_menu_names(self, sourceref):
        if self.source:
            source_name = self.source.get_name()
            submenu_label = _("Repository: %s") % source_name
        else:
            submenu_label = _("New Repository")
        return (_("Repo Reference Editor"), submenu_label)

    def ok_clicked(self, obj):
        if self.source.handle:
            # only commit if it has changed
            if self.source.serialize() != self.original:
                with DbTxn(_("Modify Repository"), self.db) as trans:
                    self.db.commit_repository(self.source, trans)
        else:
            if self.check_for_duplicate_id("Repository"):
                return
            with DbTxn(_("Add Repository"), self.db) as trans:
                self.db.add_repository(self.source, trans)
            self.source_ref.ref = self.source.handle

        if self.update:
            self.update((self.source_ref, self.source))

        self.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Import from GEDCOM"

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".GedcomImport")

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.errors import DbError, GedcomError
from gramps.plugins.lib.libmixin import DbMixin
from gramps.plugins.lib import libgedcom
from gramps.gen.utils.libformatting import ImportInfo

# The following code is necessary to ensure that when Help->Plugin
# Manager->Reload is executed, not only is the top-level exportgedcom file
# reloaded, but also the dependent libgedcom. This ensures that testing can have
# a quick turnround, without having to restart Gramps.
module = __import__(
    "gramps.plugins.lib.libgedcom", fromlist=["gramps.plugins.lib"]
)  # why o why ?? as above!
import importlib

importlib.reload(module)

from gramps.gen.config import config


# -------------------------------------------------------------------------
#
# importData
#
# -------------------------------------------------------------------------
def importData(database, filename, user):
    """
    Try to handle ANSEL encoded files that are not really ANSEL encoded
    """

    if DbMixin not in database.__class__.__bases__:
        database.__class__.__bases__ = (DbMixin,) + database.__class__.__bases__
    try:
        # Opening in utf-8 with universal newline to allow cr, lf, and crlf
        # If the file is really UTF16 or a varient, the next block code will not
        # find anything even if it is there, but this is ok since it won't be
        # ANSEL, or is inconsistent...
        with open(
            filename, "r", encoding="utf-8", errors="replace", newline=None
        ) as ifile:
            ansel = False
            gramps = False
            for index in range(50):
                # Treat the file as though it is UTF-8 since this is the more
                # modern option; and anyway it doesn't really matter as we are
                # only trying to detect a CHAR or SOUR line which is only
                # 7-bit ASCII anyway,  and we ignore anything that can't be
                # translated.
                line = ifile.readline()
                line = line.split()
                if len(line) == 0:
                    break
                if len(line) > 2 and line[1][0:4] == "CHAR" and line[2] == "ANSEL":
                    ansel = True
                if len(line) > 2 and line[1][0:4] == "SOUR" and line[2] == "GRAMPS":
                    gramps = True
    except IOError:
        return

    if not gramps and ansel and user.uistate:
        from gramps.gui.glade import Glade

        top = Glade()
        code = top.get_object("codeset")
        code.set_active(0)
        dialog = top.toplevel
        dialog.set_transient_for(user.uistate.window)
        dialog.run()
        enc = ["ANSEL", "ANSEL", "ANSI", "ASCII", "UTF-8"]
        code_set = enc[code.get_active()]
        dialog.destroy()
    else:
        code_set = ""

    assert isinstance(code_set, str)

    try:
        ifile = open(filename, "rb")
        stage_one = libgedcom.GedcomStageOne(ifile)
        stage_one.parse()

        if code_set:
            stage_one.set_encoding(code_set)
        ifile.seek(0)
        if database.get_feature("skip-import-additions"):  # don't add source or tags
            gedparse = libgedcom.GedcomParser(
                database, ifile, filename, user, stage_one, None, None
            )
        else:
            gedparse = libgedcom.GedcomParser(
                database,
                ifile,
                filename,
                user,
                stage_one,
                config.get("preferences.default-source"),
                (
                    config.get("preferences.tag-on-import-format")
                    if config.get("preferences.tag-on-import")
                    else None
                ),
            )
    except IOError as msg:
        user.notify_error(_("%s could not be opened\n") % filename, str(msg))
        return
    except GedcomError as msg:
        user.notify_error(
            _("Invalid GEDCOM file"),
            _("%s could not be imported") % filename + "\n" + str(msg),
        )
        return

    try:
        read_only = database.readonly
        database.readonly = False
        gedparse.parse_gedcom_file(False)
        database.readonly = read_only
        ifile.close()
    except IOError as msg:
        msg = _("%s could not be opened\n") % filename
        user.notify_error(msg, str(msg))
        return
    except DbError as msg:
        user.notify_db_error(str(msg.value))
        return
    except GedcomError as msg:
        user.notify_error(_("Error reading GEDCOM file"), str(msg))
        return
    ## a "GEDCOM import report" happens in GedcomParser so this is not needed:
    ## (but the imports_test.py unittest currently requires it, so here it is)
    return ImportInfo({_("Results"): _("done")})



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016 Tom Samstag
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest that tests note-specific filter rules
"""
import unittest
import os

from ....db.utils import import_as_dict
from ....filters import GenericFilterFactory
from ....const import DATA_DIR
from ....user import User

from ..note import (
    AllNotes,
    HasIdOf,
    RegExpIdOf,
    HasNote,
    MatchesRegexpOf,
    HasReferenceCountOf,
    NotePrivate,
    ChangedSince,
    HasTag,
    HasType,
)

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")
GenericNoteFilter = GenericFilterFactory("Note")


class BaseTest(unittest.TestCase):
    """
    Note rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = import_as_dict(EXAMPLE, User())

    def filter_with_rule(self, rule):
        """
        Apply a filter with the given rule.
        """
        filter_ = GenericNoteFilter()
        filter_.add_rule(rule)
        results = filter_.apply(self.db)
        return set(results)

    def test_allnotes(self):
        """
        Test AllNotes rule.
        """
        rule = AllNotes([])
        self.assertEqual(
            len(self.filter_with_rule(rule)), self.db.get_number_of_notes()
        )

    def test_hasidof(self):
        """
        Test HasIdOf rule.
        """
        rule = HasIdOf(["N0001"])
        self.assertEqual(self.filter_with_rule(rule), set(["ac380498bac48eedee8"]))

    def test_regexpidof(self):
        """
        Test RegExpIdOf rule.
        """
        rule = RegExpIdOf(["N000."], use_regex=True)
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "ac380498c020c7bcdc7",
                    "ac3804a842b21358c97",
                    "ae13613d581506d040892f88a21",
                    "ac3804a8405171ef666",
                    "ac3804a1d747a39822c",
                    "ac3804aac6b762b75a5",
                    "ac380498bac48eedee8",
                    "ac3804a1d66258b8e13",
                    "ac380498bc46102e1e8",
                    "b39fe2e143d1e599450",
                ]
            ),
        )

    def test_hasnote(self):
        """
        Test HasNote rule.
        """
        rule = HasNote(["note", "Person Note"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(["b39ff11d8912173cded", "b39ff01f75c1f76859a"]),
        )

    def test_matchesregexpof(self):
        """
        Test MatchesRegexpOf rule.
        """
        rule = MatchesRegexpOf(["^This"], use_regex=True)
        self.assertEqual(
            self.filter_with_rule(rule),
            set(["b39ff11d8912173cded", "c140d4c29520c92055c", "b39ff01f75c1f76859a"]),
        )

    def test_hasreferencecountof(self):
        """
        Test HasReferenceCountOf rule.
        """
        rule = HasReferenceCountOf(["greater than", "1"])
        self.assertEqual(self.filter_with_rule(rule), set(["c140d4c29520c92055c"]))

    def test_noteprivate(self):
        """
        Test NotePrivate rule.
        """
        rule = NotePrivate([])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_changedsince(self):
        """
        Test ChangedSince rule.
        """
        rule = ChangedSince(["2010-01-01", "2016-01-01"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "c140d4c29520c92055c",
                    "d0436bcc69d6bba278bff5bc7db",
                    "b39fe2e143d1e599450",
                    "d0436bba4ec328d3b631259a4ee",
                    "d0436be64ac277b615b79b34e72",
                ]
            ),
        )

    def test_hastag(self):
        """
        Test HasTag rule.
        """
        rule = HasTag(["ToDo"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(["b39ff01f75c1f76859a", "b39fe2e143d1e599450"]),
        )

    def test_hastype(self):
        """
        Test HasType rule.
        """
        rule = HasType(["Person Note"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(["ac380498c020c7bcdc7", "b39ff11d8912173cded", "b39ff01f75c1f76859a"]),
        )


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013 Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Unittest for object_to_data, data_to_object"""

import os
import unittest

from ...const import DATA_DIR
from ...db.utils import import_as_dict
from ...user import User
from .. import (
    Citation,
    Event,
    Family,
    Media,
    Note,
    Person,
    Place,
    Repository,
    Source,
    Tag,
)
from ..json_utils import object_to_data, data_to_object

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")


class BaseCheck:
    def test_data_to_object(self):
        data = object_to_data(self.object)
        self.assertIs(data.handle, None)
        self.assertIs(data["handle"], None)
        obj = data_to_object(data)
        self.assertEqual(id(self.object), id(obj))
        self.assertEqual(self.object.serialize(), obj.serialize())


class PersonCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Person
        self.object = self.cls()


class FamilyCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Family
        self.object = self.cls()


class EventCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Event
        self.object = self.cls()


class SourceCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Source
        self.object = self.cls()


class PlaceCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Place
        self.object = self.cls()


class CitationCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Citation
        self.object = self.cls()


class RepositoryCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Repository
        self.object = self.cls()


class MediaCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Media
        self.object = self.cls()


class NoteCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Note
        self.object = self.cls()


class TagCheck(unittest.TestCase, BaseCheck):
    def setUp(self):
        self.cls = Tag
        self.object = self.cls()


class DatabaseCheck(unittest.TestCase):
    maxDiff = None


def generate_cases(obj, data):
    """
    Dynamically generate tests and attach to DatabaseCheck.
    """

    def test(self):
        json_data = object_to_data(obj)
        obj2 = data_to_object(json_data)
        self.assertEqual(obj.serialize(), obj2.serialize())

    name = "test_serialize_%s_%s" % (obj.__class__.__name__, obj.handle)
    setattr(DatabaseCheck, name, test)

    def test_data(self):
        class_name = obj.__class__.__name__
        self.assertIsInstance(data, dict, "Ensure that the data is a dict")
        self.assertEqual(data.handle, data["handle"], "Test attribute access")

        if class_name == "Person":
            if (len(data.parent_family_list)) > 0:
                # Get a handle:
                self.assertIsInstance(
                    data.parent_family_list[0], str, "Test list access"
                )

        self.assertNotIn("_object", data.keys(), "Object not created")
        self.assertEqual(data.get_handle(), data["handle"], "Test method call")
        self.assertIn("_object", data.keys(), "Object created")
        self.assertEqual(data["_object"].handle, data["handle"], "Object is correct")
        self.assertEqual(
            data["_object"].__class__.__name__, class_name, "Object is correct type"
        )

    name = "test_data_%s_%s" % (obj.__class__.__name__, obj.handle)
    setattr(DatabaseCheck, name, test_data)


db = import_as_dict(EXAMPLE, User())
for obj_class in (
    "Person",
    "Family",
    "Event",
    "Place",
    "Repository",
    "Source",
    "Citation",
    "Media",
    "Note",
):
    for handle in db.method("get_%s_handles", obj_class)():
        obj = db.method("get_%s_from_handle", obj_class)(handle)
        data = db.method("get_raw_%s_data", obj_class)(handle)
        generate_cases(obj, data)

if __name__ == "__main__":
    unittest.main()



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.widgets import Photo
from gramps.gen.utils.file import media_path_full


class MediaPreview(Gramplet):
    """
    Displays a preview of the media object.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.top = Gtk.Box()
        self.photo = Photo(self.uistate.screen_height() < 1000)
        self.top.pack_start(self.photo, fill=True, expand=False, padding=5)
        self.top.show_all()
        return self.top

    def db_changed(self):
        self.connect(self.dbstate.db, "media-update", self.update)
        self.connect_signal("Media", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Media")
        if active_handle:
            active_media = self.dbstate.db.get_media_from_handle(active_handle)
            self.set_has_data(active_media is not None)
        else:
            self.set_has_data(False)

    def main(self):
        active_handle = self.get_active("Media")
        if active_handle:
            media = self.dbstate.db.get_media_from_handle(active_handle)
            self.top.hide()
            if media:
                self.load_image(media)
                self.set_has_data(True)
            else:
                self.photo.set_image(None)
                self.set_has_data(False)
            self.top.show()
        else:
            self.photo.set_image(None)
            self.set_has_data(False)

    def load_image(self, media):
        """
        Load the primary image if it exists.
        """
        self.full_path = media_path_full(self.dbstate.db, media.get_path())
        mime_type = media.get_mime_type()
        self.photo.set_image(self.full_path, mime_type)
        self.photo.set_uistate(self.uistate, None)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
File and folder related utility functions
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os
import sys
import shutil
import tempfile
import hashlib
import logging

LOG = logging.getLogger(".gen.utils.file")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..constfunc import win, mac, get_env_var
from ..const import USER_HOME, USER_PICTURES, ENV, GRAMPS_LOCALE as glocale

# -------------------------------------------------------------------------
#
#  Constants
#
# -------------------------------------------------------------------------
_NEW_NAME_PATTERN = "%s%sUntitled_%d.%s"


# -------------------------------------------------------------------------
#
#  Functions
#
# -------------------------------------------------------------------------
def find_file(filename):
    # try the filename we got
    try:
        if os.path.isfile(filename):
            return filename
    except UnicodeError as err:
        LOG.error("Filename %s raised a Unicode Error %s.", repr(filename), err)

    LOG.debug("Filename %s not found.", repr(filename))
    return ""


def find_folder(filename):
    # try the filename we got
    try:
        if os.path.isdir(filename):
            return filename
    except UnicodeError as err:
        LOG.error("Filename %s raised a Unicode Error %s", repr(filename), err)

    LOG.debug("Filename %s either not found or not a directory.", repr(filename))
    return ""


def get_new_filename(ext, folder="~/"):
    ix = 1
    while os.path.isfile(
        os.path.expanduser(_NEW_NAME_PATTERN % (folder, os.path.sep, ix, ext))
    ):
        ix = ix + 1
    return os.path.expanduser(_NEW_NAME_PATTERN % (folder, os.path.sep, ix, ext))


def get_empty_tempdir(dirname):
    """Return path to TEMP_DIR/dirname, a guaranteed empty directory

    makes intervening directories if required
    fails if _file_ by that name already exists,
    or for inadequate permissions to delete dir/files or create dir(s)

    """
    dirpath = tempfile.mkdtemp(prefix="gramps-", suffix="-" + dirname)
    return dirpath


def rm_tempdir(path):
    """Remove a tempdir created with get_empty_tempdir"""
    if path.startswith(tempfile.gettempdir()) and os.path.isdir(path):
        shutil.rmtree(path)


def relative_path(original, base):
    """
    Calculate the relative path from base to original, with base a directory,
    and original an absolute path
    On problems, original is returned unchanged
    """
    if not os.path.isdir(base):
        return original
    # original and base must be absolute paths
    if not os.path.isabs(base):
        return original
    if not os.path.isabs(original):
        return original
    original = os.path.normpath(original)
    base = os.path.normpath(base)

    # If the db_dir and obj_dir are on different drives (win only)
    # then there cannot be a relative path. Return original obj_path
    (base_drive, base) = os.path.splitdrive(base)
    (orig_drive, orig_name) = os.path.splitdrive(original)
    if base_drive.upper() != orig_drive.upper():
        return original

    # Starting from the filepath root, work out how much of the filepath is
    # shared by base and target.
    base_list = (base).split(os.sep)
    target_list = (orig_name).split(os.sep)
    # make sure '/home/person' and 'c:/home/person' both give
    #   list ['home', 'person']
    base_list = [_f for _f in base_list if _f]
    target_list = [_f for _f in target_list if _f]
    i = -1
    # base path is normcase (lower case on Windows) so compare target in lower
    # on Windows as well
    for i in range(min(len(base_list), len(target_list))):
        if win():
            if base_list[i].lower() != target_list[i].lower():
                break
        else:
            if base_list[i] != target_list[i]:
                break
    else:
        # if break did not happen we are here at end, and add 1.
        i += 1
    rel_list = [os.pardir] * (len(base_list) - i) + target_list[i:]
    return os.path.join(*rel_list)


def expand_path(path, normalize=True):
    """
    Expand environment variables in a path
    Uses both the environment variables and the Gramps environment
    The expansion uses the str.format, e.g. "~/{GRAMPSHOME}/{VERSION}/filename.txt"
    We make the assumption that the user will not use a path that contain variable names
    (it is technically possible to use characters "{", "}" in  paths)
    """
    environment = dict(os.environ)
    environment.update(ENV)
    if not "GRAMPSHOME" in environment:
        environment["GRAMPSHOME"] = USER_HOME
    path = path.format(**environment)
    if normalize:
        path = os.path.normcase(os.path.normpath(os.path.abspath(path)))
    return path


def media_path(db):
    """
    Given a database, return the mediapath to use as basedir for media
    """
    mpath = db.get_mediapath()
    return expand_media_path(mpath, db)


def expand_media_path(mpath, db):
    """
    Normalize a mediapath:
     - Relative mediapath are considered as relative to the database
     - Expand variables, see expand_path
     - Convert to absolute path
     - Convert slashes and case (on Windows)
    """
    # Use XDG pictures diectory if no media_path specified
    if mpath is None:
        mpath = os.path.abspath(USER_PICTURES)
    # Expand environment variables
    mpath = expand_path(mpath, False)
    # Relative mediapath are considered as relative to the database
    if not os.path.isabs(mpath):
        basepath = db.get_save_path()
        if not basepath:
            basepath = USER_HOME
        mpath = os.path.join(os.path.abspath(basepath), mpath)
    # Normalize path
    mpath = os.path.normcase(os.path.normpath(os.path.abspath(mpath)))
    return mpath


def media_path_full(db, filename):
    """
    Given a database and a filename of a media, return the media filename
    is full form, eg 'graves/tomb.png' becomes '/home/me/genea/graves/tomb.png
    """
    if filename.startswith(("http://", "https://")):
        return filename
    if os.path.isabs(filename):
        return filename
    mpath = media_path(db)
    return os.path.join(mpath, filename)


def search_for(name):
    if name.startswith('"'):
        name = name.split('"')[1]
    else:
        name = name.split()[0]
    if win():
        for i in get_env_var("PATH").split(";"):
            fname = os.path.join(i, name)
            if os.access(fname, os.X_OK) and not os.path.isdir(fname):
                return 1
        if os.access(name, os.X_OK) and not os.path.isdir(name):
            return 1
    else:
        for i in os.environ["PATH"].split(":"):  # not win()
            fname = os.path.join(i, name)
            if os.access(fname, os.X_OK) and not os.path.isdir(fname):
                return 1
    return 0


def where_is(name):
    """This command is similar to the Linux "whereis -b file" command.
    It looks for an executable file (name) in the PATH python is using, as
    well as several likely other paths.  It returns the first file found,
    or an empty string if not found.
    """
    paths = set(os.environ["PATH"].split(os.pathsep))
    if not win():
        paths.update(
            ("/bin", "/usr/bin", "/usr/local/bin", "/opt/local/bin", "/opt/bin")
        )
    for i in paths:
        fname = os.path.join(i, name)
        if os.access(fname, os.X_OK) and not os.path.isdir(fname):
            return fname
    return ""


def create_checksum(full_path):
    """
    Create a md5 hash for the given file.
    """
    full_path = os.path.normpath(full_path)
    md5 = hashlib.md5()
    try:
        with open(full_path, "rb") as media_file:
            while True:
                buf = media_file.read(65536)
                if not buf:
                    break
                md5.update(buf)
        md5sum = md5.hexdigest()
    except IOError:
        md5sum = ""
    except UnicodeEncodeError:
        md5sum = ""
    return md5sum



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007 Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Unittest for testing ..."""

import unittest

from ..keyword import (
    KEYWORDS,
    get_translation_from_keyword,
    get_keyword_from_translation,
)


class TestCase(unittest.TestCase):
    def keyword_case(self, item1, item2):
        result = get_translation_from_keyword(item1)
        self.assertEqual(result, item2)

    def translation_case(self, item1, item2):
        result = get_keyword_from_translation(item1)
        self.assertEqual(result, item2)

    def test_from_keyword(self):
        for keyword, code, standard, upper in KEYWORDS:
            self.keyword_case(keyword, standard)

    def test_from_translation(self):
        for keyword, code, standard, upper in KEYWORDS:
            self.translation_case(standard, keyword)

    def test_from_lower(self):
        for keyword, code, standard, upper in KEYWORDS:
            self.translation_case(standard.lower(), keyword)

    def test_from_upper(self):
        for keyword, code, standard, upper in KEYWORDS:
            self.translation_case(upper, keyword.upper())


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Note
from ....db import Database


# -------------------------------------------------------------------------
# "Events having notes that contain a substring"
# -------------------------------------------------------------------------
class MatchesSubstringOf(Rule):
    """Notes having notes containing <substring>"""

    labels = [_("Substring:")]
    name = _("Notes containing <substring>")
    description = _("Matches notes that contain text " "which matches a substring")
    category = _("General filters")

    def apply_to_one(self, db: Database, note: Note) -> bool:
        """Apply the filter"""
        text = str(note.text)
        if text.upper().find(self.list[0].upper()) != -1:
            return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
DashboardView interface.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.views.pageview import PageView
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gui.widgets.grampletpane import GrampletPane


class DashboardView(PageView):
    """
    DashboardView interface
    """

    def __init__(self, pdata, dbstate, uistate):
        """
        Create a DashboardView, with the current dbstate and uistate
        """
        PageView.__init__(self, _("Dashboard"), pdata, dbstate, uistate)
        self.ui_def = []  # No special menu for Dashboard, Popup in GrampletPane

    def build_interface(self):
        """
        Builds the container widget for the interface.
        Returns a gtk container widget.
        """
        top = self.build_widget()
        top.show_all()
        return top

    def build_widget(self):
        """
        Builds the container widget for the interface. Must be overridden by the
        the base class. Returns a gtk container widget.
        """
        # load the user's gramplets and set columns, etc
        self.widget = GrampletPane(
            "Gramplets_dashboardview_gramplets", self, self.dbstate, self.uistate
        )
        return self.widget

    def build_tree(self):
        """
        Rebuilds the current display.
        """
        pass

    def get_title(self):
        """
        Used to set the titlebar in the configuration window.
        """
        return _("Dashboard")

    def get_stock(self):
        """
        Return image associated with the view, which is used for the
        icon for the button.
        """
        return "gramps-gramplet"

    def get_viewtype_stock(self):
        """Type of view in category"""
        return "gramps-gramplet"

    def define_actions(self):
        """
        Defines the UIManager actions. Called by the ViewManager to set up the
        View. The user typically defines self.action_list and
        self.action_toggle_list in this function.
        """
        pass

    def set_inactive(self):
        self.active = False
        self.widget.set_inactive()

    def set_active(self):
        new_title = "%s - %s - Gramps" % (
            self.dbstate.db.get_dbname(),
            self.get_title(),
        )
        self.uistate.window.set_title(new_title)
        self.active = True
        self.widget.set_active()

    def on_delete(self):
        self.widget.on_delete()
        self._config.save()

    def can_configure(self):
        """
        See :class:`~gui.views.pageview.PageView
        :return: bool
        """
        return self.widget.can_configure()

    def _get_configure_page_funcs(self):
        """
        Return a list of functions that create gtk elements to use in the
        notebook pages of the Configure dialog

        :return: list of functions
        """
        return self.widget._get_configure_page_funcs()

    def navigation_type(self):
        """
        Return a description of the specific nav_type items that are
        associated with this view. None means that there is no specific
        type.
        """
        return None



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Robert Ham <rah@bash.sh>
# Copyright (C) 2011 Adam Stein <adam@csh.rit.edu>
# Copyright (C) 2017 Paul Culley <paulr2787@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
from collections import deque

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from . import MatchesFilter
from ....const import GRAMPS_LOCALE as glocale

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Union, List, Set, Dict
from ....lib import Person
from ....db import Database
from ....types import FamilyHandle, PersonHandle

_ = glocale.translation.gettext
# -------------------------------------------------------------------------
#
# DeepRelationshipPathBetween
#
# -------------------------------------------------------------------------


def get_family_handle_people(
    db: Database, exclude_handle: str, family_handle: FamilyHandle
):
    people: Set[str] = set()

    family = db.get_family_from_handle(family_handle)

    def possibly_add_handle(h: str):
        if h is not None and h != exclude_handle:
            people.add(h)

    possibly_add_handle(family.father_handle)
    possibly_add_handle(family.mother_handle)

    for child_ref in family.child_ref_list:
        if child_ref:
            possibly_add_handle(child_ref.ref)

    return people


def get_person_family_people(
    db: Database, person: Person, person_handle: PersonHandle
) -> Set[str]:
    people: Set[str] = set()

    def add_family_handle_list(fam_list: List[FamilyHandle]):
        for family_handle in fam_list:
            people.update(get_family_handle_people(db, person_handle, family_handle))

    add_family_handle_list(person.family_list)
    add_family_handle_list(person.parent_family_list)

    return people


def find_deep_relations(
    db: Database, user, person: Person | None, target_people: List[str]
) -> Set[str]:
    """This explores all possible paths between a person and one or more
    targets.  The algorithm processes paths in a breadth first wave, one
    remove at a time.  The first path that reaches a target causes the target
    to be marked complete and the path to be stored in the return_paths set.
    By processing in wave, the return path should be a shortest path.
    The function stores to do data and intermediate results in an ordered dict,
    rather than using a recursive algorithm because some trees have been found
    that exceed the standard python recursive depth."""
    return_paths: Set[str] = set()  # all people in paths between targets and person
    if person is None:
        return return_paths
    todo = deque([person.handle])  # list of work to do, handles, add to right,
    #                                pop from left
    done: Dict[str, Union[str, None]] = {}  # The key records handles already examined,
    # the value is a handle of the previous person in the path, or None at
    # head of path.  This forms a linked list of handles along the path.
    done[person.handle] = None

    while todo:
        handle = todo.popleft()

        if user:
            user.step_progress()

        if handle in target_people:  # if we found a target
            prev_hndl = handle
            # Go through linked list and save the handles in return_paths
            while prev_hndl:
                return_paths.add(prev_hndl)
                prev_hndl = done[prev_hndl]
            target_people.remove(handle)
            if not target_people:  # Quit searching if all targets found
                break

        person = db.get_person_from_handle(handle)
        if person is None:
            continue

        people = get_person_family_people(db, person, handle)
        for p_hndl in people:
            if p_hndl in done:  # check if we have already been here
                continue  # and ignore if we have
            todo.append(p_hndl)  # Add to the todo list
            done[p_hndl] = handle  # Add to the (almost) done list

    return return_paths


class DeepRelationshipPathBetween(Rule):
    """Checks if there is any familial connection between a person and a
    filter match by searching over all connections."""

    labels = [_("ID:"), _("Filter name:")]
    name = _("Relationship path between <person> and people matching <filter>")
    category = _("Relationship filters")
    description = _(
        "Searches over the database starting from a specified"
        " person and returns everyone between that person and"
        " a set of target people specified with a filter.  "
        "This produces a set of relationship paths (including"
        " by marriage) between the specified person and the"
        " target people.  Each path is not necessarily"
        " the shortest path."
    )

    def prepare(self, db: Database, user):
        root_person_id = self.list[0]
        root_person = db.get_person_from_gramps_id(root_person_id)

        filter_name = self.list[1]
        self.filt = MatchesFilter([filter_name])
        self.filt.requestprepare(db, user)

        if user:
            user.begin_progress(
                _("Finding relationship paths"),
                _("Retrieving all sub-filter matches"),
                db.get_number_of_people(),
            )
        target_people = []
        for person in db.iter_people():
            if self.filt.apply_to_one(db, person):
                target_people.append(person.handle)
            if user:
                user.step_progress()
        if user:
            user.end_progress()
            user.begin_progress(
                _("Finding relationship paths"),
                _("Evaluating people"),
                db.get_number_of_people(),
            )
        self.selected_handles: Set[str] = find_deep_relations(
            db, user, root_person, target_people
        )
        if user:
            user.end_progress()

    def reset(self):
        self.filt.requestreset()
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles



#! /usr/bin/env python3
#
# update_po - a gramps tool to update translations
#
# Copyright (C) 2006-2006  Kees Bakker
# Copyright (C) 2006       Brian Matherly
# Copyright (C) 2008       Stephen George
# Copyright (C) 2012
# Copyright (C) 2020       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
update_po.py for Gramps translations.

Examples:
   python update_po.py -t

      Tests if 'gettext' and 'python' are well configured.

   python update_po.py -h

      Calls help and command line interface.

   python update_po.py -p

      Generates a new template/catalog (gramps.pot).

   python update_po.py -m de.po

      Merges 'de.po' file with 'gramps.pot'.

   python update_po.py -k de.po

      Checks 'de.po' file, tests to compile and generates a textual resume.
"""

from __future__ import print_function

import os
import sys
import shutil
from argparse import ArgumentParser
from tokenize import tokenize, STRING, COMMENT, NL, TokenError

# Windows OS

if sys.platform in ["linux", "linux2", "darwin", "cygwin"] or shutil.which("msgmerge"):
    msgmergeCmd = "msgmerge"
    msgfmtCmd = "msgfmt"
    msgattribCmd = "msgattrib"
    xgettextCmd = "xgettext"
    pythonCmd = os.path.join(sys.prefix, "bin", "python3")
elif sys.platform == "win32":
    # GetText Win 32 obtained from http://gnuwin32.sourceforge.net/packages/gettext.htm
    # ....\gettext\bin\msgmerge.exe needs to be on the path
    msgmergeCmd = os.path.join(
        "C:", "Program Files(x86)", "gettext", "bin", "msgmerge.exe"
    )
    msgfmtCmd = os.path.join("C:", "Program Files(x86)", "gettext", "bin", "msgfmt.exe")
    msgattribCmd = os.path.join(
        "C:", "Program Files(x86)", "gettext", "bin", "msgattrib.exe"
    )
    xgettextCmd = os.path.join(
        "C:", "Program Files(x86)", "gettext", "bin", "xgettext.exe"
    )
    pythonCmd = os.path.join(sys.prefix, "bin", "python.exe")

# Others OS

elif sys.platform in ["linux", "linux2", "darwin", "cygwin"]:
    msgmergeCmd = "msgmerge"
    msgfmtCmd = "msgfmt"
    msgattribCmd = "msgattrib"
    xgettextCmd = "xgettext"
    pythonCmd = os.path.join(sys.prefix, "bin", "python3")
else:
    print("Found platform %s, OS %s" % (sys.platform, os.name))
    print("Update PO ERROR: unknown system, don't know msgmerge, ... commands")
    sys.exit(0)

# List of available languages, useful for grouped actions

# need files with po extension
LANG = [file for file in os.listdir(".") if file.endswith(".po")]
# add a special 'all' argument (for 'check' and 'merge' arguments)
LANG.append("all")
# visual polish on the languages list
LANG.sort()


def tests():
    """
    Testing installed programs.
    We made tests (-t flag) by displaying versions of tools if properly
    installed. Cannot run all commands without 'gettext' and 'python'.
    """
    try:
        print("\n====='msgmerge'=(merge our translation)================\n")
        os.system("""%(program)s -V""" % {"program": msgmergeCmd})
    except:
        print(
            "Please, install %(program)s for updating your translation"
            % {"program": msgmergeCmd}
        )

    try:
        print("\n==='msgfmt'=(format our translation for installation)==\n")
        os.system("""%(program)s -V""" % {"program": msgfmtCmd})
    except:
        print(
            "Please, install %(program)s for checking your translation"
            % {"program": msgfmtCmd}
        )

    try:
        print("\n===='msgattrib'==(list groups of messages)=============\n")
        os.system("""%(program)s -V""" % {"program": msgattribCmd})
    except:
        print(
            "Please, install %(program)s for listing groups of messages"
            % {"program": msgattribCmd}
        )

    try:
        print("\n===='xgettext' =(generate a new template)==============\n")
        os.system("""%(program)s -V""" % {"program": xgettextCmd})
    except:
        print(
            "Please, install %(program)s for generating a new template"
            % {"program": xgettextCmd}
        )

    try:
        print("\n=================='python'=============================\n")
        os.system("""%(program)s -V""" % {"program": pythonCmd})
    except:
        print("Please, install python")


def main():
    """
    The utility for handling translation stuff.
    What is need by Gramps, nothing more.
    """

    parser = ArgumentParser(
        description="This program generates a new template and "
        "also provides some common features.",
    )
    parser.add_argument(
        "-t",
        "--test",
        action="store_true",
        dest="test",
        default=True,
        help="test if 'python' and 'gettext' are properly installed",
    )

    parser.add_argument(
        "-x",
        "--xml",
        action="store_true",
        dest="xml",
        default=False,
        help="extract messages from xml based file formats",
    )
    parser.add_argument(
        "-g",
        "--glade",
        action="store_true",
        dest="glade",
        default=False,
        help="extract messages from glade file format only",
    )
    parser.add_argument(
        "-c",
        "--clean",
        action="store_true",
        dest="clean",
        default=False,
        help="remove created files",
    )
    parser.add_argument(
        "-p",
        "--pot",
        action="store_true",
        dest="catalog",
        default=False,
        help="create a new catalog",
    )

    update = parser.add_argument_group("Update", "Maintenance around translations")

    # need at least one argument (sv.po, de.po, etc ...)

    # lang.po files maintenance
    update.add_argument(
        "-m", dest="merge", choices=LANG, help="merge lang.po files with last catalog"
    )

    update.add_argument("-k", dest="check", choices=LANG, help="check lang.po files")

    # testing stage
    trans = parser.add_argument_group(
        "Translation", "Display content of translations file"
    )

    # need one argument (eg, de.po)

    trans.add_argument(
        "-u",
        dest="untranslated",
        choices=[file for file in os.listdir(".") if file.endswith(".po")],
        help="list untranslated messages",
    )
    trans.add_argument(
        "-f",
        dest="fuzzy",
        choices=[file for file in os.listdir(".") if file.endswith(".po")],
        help="list fuzzy messages",
    )

    args = parser.parse_args()
    namespace, extra = parser.parse_known_args()

    if args.test:
        tests()

    if args.xml:
        extract_xml()

    if args.glade:
        create_filesfile()
        extract_glade()
        if os.path.isfile("tmpfiles"):
            os.unlink("tmpfiles")

    if args.catalog:
        retrieve()

    if args.clean:
        clean()

    if args.merge:
        # retrieve() windows os?
        if sys.argv[2:] == ["all"]:
            sys.argv[2:] = LANG
        merge(sys.argv[2:])

    if args.check:
        # retrieve() windows os?
        if sys.argv[2:] == ["all"]:
            sys.argv[2:] = LANG
        check(sys.argv[2:])

    if args.untranslated:
        untranslated(sys.argv[2:])

    if args.fuzzy:
        fuzzy(sys.argv[2:])


def create_filesfile():
    """
    Create a file with all files that we should translate.
    These are all python files not in POTFILES.skip added with those in
    POTFILES.in
    """
    dir = os.getcwd()
    topdir = os.path.normpath(os.path.join(dir, "..", "gramps"))
    lentopdir = len(topdir)
    with open("POTFILES.in") as f:
        infiles = dict(
            ["../" + file.strip(), None]
            for file in f
            if file.strip() and not file[0] == "#"
        )

    with open("POTFILES.skip") as f:
        notinfiles = dict(
            ["../" + file.strip(), None] for file in f if file and not file[0] == "#"
        )

    for dirpath, dirnames, filenames in os.walk(topdir):
        root, subdir = os.path.split(dirpath)
        if subdir.startswith("."):
            # don't continue in this dir
            dirnames[:] = []
            continue
        for dirname in dirnames:
            # Skip hidden and system directories:
            if dirname.startswith(".") or dirname in ["po", "locale"]:
                dirnames.remove(dirname)
        # add the files which are python or glade files
        # if the directory does not exist or is a link, do nothing
        if not os.path.isdir(dirpath) or os.path.islink(dirpath):
            continue

        for filename in os.listdir(dirpath):
            name = os.path.split(filename)[1]
            if name.endswith((".py", ".glade")):
                full_filename = os.path.join(dirpath, filename)
                # Skip the file if in POTFILES.skip
                if full_filename[lentopdir:] in notinfiles:
                    infiles["../gramps" + full_filename[lentopdir:]] = None
    # now we write out all the files in form ../gramps/filename
    with open("tmpfiles", "w") as f:
        for file in sorted(infiles.keys()):
            f.write(file)
            f.write("\n")


def listing(name, extensionlist):
    """
    List files according to extensions.
    Parsing from a textual file (gramps) is faster and easy for maintenance.
    Like POTFILES.in and POTFILES.skip
    """

    with open("tmpfiles") as f:
        files = [file.strip() for file in f if file and not file[0] == "#"]

    with open(name, "w") as temp:
        for entry in files:
            for ext in extensionlist:
                if entry.endswith(ext):
                    temp.write(entry)
                    temp.write("\n")
                    break


def headers():
    """
    Look at existing C file format headers.
    Generated by 'intltool-extract' but want to get rid of this
    dependency (perl, just a set of tools).
    """
    headers = []

    # in.h; extract_xml
    if os.path.isfile("""fragments.pot"""):
        headers.append("""fragments.pot""")

    return headers


def extract_xml():
    """
    Extract translation strings from XML based, mime and desktop files.
    Uses custom ITS rules found in the po/its directory.
    """
    if not os.path.isfile("gramps.pot"):
        create_template()

    for input_file in [
        "../data/holidays.xml",
        "../data/tips.xml",
        "../data/org.gramps_project.Gramps.xml.in",
        "../data/org.gramps_project.Gramps.metainfo.xml.in",
        "../data/org.gramps_project.Gramps.desktop.in",
    ]:
        os.system(
            (
                "GETTEXTDATADIR=. %(xgettext)s -F -j "
                "-o gramps.pot --from-code=UTF-8 %(inputfile)s"
            )
            % {"xgettext": xgettextCmd, "inputfile": input_file}
        )
        print(input_file)


def create_template():
    """
    Create a new file for template, if it does not exist.
    """
    with open("gramps.pot", "w") as template:
        pass


def extract_glade():
    """
    Extract messages from a temp file with all .glade
    """
    if not os.path.isfile("gramps.pot"):
        create_template()

    listing("glade.txt", [".glade"])
    os.system(
        """%(xgettext)s -F --add-comments -j -L Glade """
        """--from-code=UTF-8 -o gramps.pot --files-from=glade.txt"""
        % {"xgettext": xgettextCmd}
    )


def xml_fragments():
    """search through the file for xml fragments that contain the
    'translate="yes">string<' pattern.  These need to be added to the message
    catalog"""
    with open("tmpfiles") as __f:
        files = [
            file.strip()
            for file in __f
            if file and not (file[0] == "#") and file.endswith(".py\n")
        ]
        print("Checking for XML fragments in Python files")
        modop = int(len(files) / 20)
    wfp = open("fragments.pot", "w", encoding="utf-8")
    wfp.write('msgid ""\n')
    wfp.write('msgstr ""\n')
    wfp.write('"Content-Type: text/plain; charset=UTF-8\\n"\n\n')
    for indx, filename in enumerate(files):
        if not indx % modop:
            print(int(indx / len(files) * 100), end="\r")
        fp = open(filename, "rb")
        try:
            tokens = tokenize(fp.readline)
            in_string = False
            for _token, _text, _start, _end, _line in tokens:
                if _text.startswith(('"""', "'''")):
                    _text = _text[3:]
                elif _text.startswith(('"', "'")):
                    _text = _text[1:]
                if _text.endswith(('"""', "'''")):
                    _text = _text[:-3]
                elif _text.endswith(('"', "'")):
                    _text = _text[:-1]
                if _token == STRING and not in_string:
                    in_string = True
                    line_no = _start[0]
                    text = _text
                    continue
                elif _token == STRING and in_string:
                    text += _text
                    continue
                elif _token == COMMENT or _token == NL and in_string:
                    # need to ignore comments and concatinate strings
                    _ml = True
                    continue
                elif in_string:
                    in_string = False
                    end = 0
                    # _find_message_in_xml(text)
                    while True:
                        fnd = text.find('translatable="yes">', end)
                        if fnd == -1:
                            break
                        end = text.find("<", fnd)
                        if end == -1:
                            print(
                                "\nBad xml fragment '%s' at %s line %d"
                                % (text[fnd:], filename, _start[0])
                            )
                            break
                        msg = text[fnd + 19 : end]
                        if "%s" in msg or (msg.startswith("{") and msg.endswith("}")):
                            print(
                                "\n#: %s:%d  Are you sure you want to "
                                'translate the "%%s"???' % (filename, line_no)
                            )
                            break
                        wfp.write(
                            '#: %s:%d\nmsgid "%s"\nmsgstr ""\n'
                            % (filename, line_no, msg)
                        )
        except TokenError as e:
            print(
                "\n%s: %s, line %d, column %d"
                % (e.args[0], filename, e.args[1][0], e.args[1][1]),
                file=sys.stderr,
            )
        finally:
            fp.close()
    wfp.close()


def retrieve():
    """
    Extract messages from all files used by Gramps (python, glade, xml)
    """
    create_template()

    create_filesfile()
    xml_fragments()

    listing("python.txt", [".py", ".py.in"])

    # additional keywords must always be kept in sync with those in genpot.sh
    os.system(
        """%(xgettext)s -F --add-comments=Translators -j """
        """--directory=./ -d gramps -L Python """
        """-o gramps.pot --files-from=python.txt --debug """
        """--keyword=_ --keyword=_:1,2c """
        """--keyword=_T_ --keyword=_T_:1,2c """
        """--keyword=trans_text --keyword=trans_text:1,2c """
        """--keyword=ngettext --keyword=sgettext """
        """--from-code=UTF-8""" % {"xgettext": xgettextCmd}
    )

    extract_glade()
    extract_xml()

    # C format header (.h extension)
    for h in headers():
        print("xgettext for %s" % h)
        os.system(
            """%(xgettext)s -F --add-comments=Translators -j """
            """-o gramps.pot --keyword=N_ --from-code=UTF-8 %(head)s"""
            % {"xgettext": xgettextCmd, "head": h}
        )
    clean()


def clean():
    """
    Remove created files (C format headers, temp listings)
    """
    for h in headers():
        if os.path.isfile(h):
            os.unlink(h)
            print("Remove %(head)s" % {"head": h})

    if os.path.isfile("python.txt"):
        os.unlink("python.txt")
        print("Remove 'python.txt'")

    if os.path.isfile("glade.txt"):
        os.unlink("glade.txt")
        print("Remove 'glade.txt'")

    if os.path.isfile("tmpfiles"):
        os.unlink("tmpfiles")
        print("Remove 'tmpfiles'")


def merge(args):
    """
    Merge messages with 'gramps.pot'
    """
    for arg in args:
        if arg == "all":
            continue
        print("Merge %(lang)s with current template" % {"lang": arg})
        os.system(
            """%(msgmerge)s -U %(lang)s gramps.pot"""
            % {"msgmerge": msgmergeCmd, "lang": arg}
        )
        print("Updated file: '%(lang)s'." % {"lang": arg})


def check(args):
    """
    Check the translation file
    """
    for arg in args:
        if arg == "all":
            continue
        print(
            "Checked file: '%(lang.po)s'. See '%(txt)s.txt'."
            % {"lang.po": arg, "txt": arg[:-3]}
        )
        os.system(
            """%(python)s ./check_po -s %(lang.po)s > %(lang)s.txt"""
            % {"python": pythonCmd, "lang.po": arg, "lang": arg[:-3]}
        )
        os.system(
            """%(msgfmt)s -c -v %(lang.po)s""" % {"msgfmt": msgfmtCmd, "lang.po": arg}
        )


def untranslated(arg):
    """
    List untranslated messages
    """
    os.system(
        """%(msgattrib)s --untranslated %(lang.po)s"""
        % {"msgattrib": msgattribCmd, "lang.po": arg[0]}
    )


def fuzzy(arg):
    """
    List fuzzy messages
    """
    os.system(
        """%(msgattrib)s --only-fuzzy --no-obsolete %(lang.po)s"""
        % {"msgattrib": msgattribCmd, "lang.po": arg[0]}
    )


if __name__ == "__main__":
    main()



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016  Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Geography for one family
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import operator

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import EventRoleType, EventType
from gramps.gen.config import config
from gramps.gen.datehandler import displayer
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd
from gramps.gen.utils.place import conv_lat_lon
from gramps.gui.views.bookmarks import FamilyBookmarks
from gramps.plugins.lib.maps.geography import GeoGraphyView

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
_LOG = logging.getLogger("GeoGraphy.geofamily")
KEY_TAB = Gdk.KEY_Tab
_ = glocale.translation.gettext

_UI_DEF = [
    """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
    """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.PrintView</attribute>
          <attribute name="label" translatable="yes">Print...</attribute>
        </item>
      </section>
""",
    """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
    % _("Organize Bookmarks"),  # Following are the Toolbar items
    """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
     </placeholder>
""",
    """
    <placeholder id='BarCommonEdit'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">document-print</property>
        <property name="action-name">win.PrintView</property>
        <property name="tooltip_text" translatable="yes">"""
    """Print or save the Map</property>
        <property name="label" translatable="yes">Print...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
    """,
]

# pylint: disable=no-member
# pylint: disable=unused-variable
# pylint: disable=unused-argument


# -------------------------------------------------------------------------
#
# GeoView
#
# -------------------------------------------------------------------------
class GeoFamily(GeoGraphyView):
    """
    The view used to render family map.
    """

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        GeoGraphyView.__init__(
            self,
            _("Family places map"),
            pdata,
            dbstate,
            uistate,
            FamilyBookmarks,
            nav_group,
        )
        self.dbstate = dbstate
        self.uistate = uistate
        self.place_list = []
        self.place_without_coordinates = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        self.nbplaces = 0
        self.nbmarkers = 0
        self.sort = []
        self.additional_uis.append(self.additional_ui())
        self.no_show_places_in_status_bar = False
        self.cal = None

    def get_title(self):
        """
        Used to set the titlebar in the configuration window.
        """
        return _("GeoFamily")

    def get_stock(self):
        """
        Returns the name of the stock icon to use for the display.
        This assumes that this icon has already been registered
        as a stock icon.
        """
        return "geo-show-family"

    def get_viewtype_stock(self):
        """Type of view in category"""
        return "geo-show-family"

    def additional_ui(self):
        """
        Specifies the UIManager XML code that defines the menus and buttons
        associated with the interface.
        """
        return _UI_DEF

    def navigation_type(self):
        """
        Indicates the navigation type. Navigation type can be the string
        name of any of the primary objects.
        """
        return "Family"

    def goto_handle(self, handle=None):
        """
        Rebuild the tree with the given person handle as the root.
        """
        if self.osm is None:
            return
        if self.uistate.get_active("Family"):
            self._createmap(self.uistate.get_active("Family"))
        else:
            self._createmap(self.uistate.get_active("Person"))

    def build_tree(self):
        """
        This is called by the parent class when the view becomes visible. Since
        all handling of visibility is now in rebuild_trees, see that for more
        information.
        """
        pass

    def _createpersonmarkers(self, dbstate, person, comment, fam_id):
        """
        Create all markers for the specified person.
        """
        self.cal = config.get("preferences.calendar-format-report")
        latitude = longitude = ""
        if person:
            # For each event, if we have a place, set a marker.
            for event_ref in person.get_event_ref_list():
                if not event_ref:
                    continue
                role = event_ref.get_role()
                event = dbstate.db.get_event_from_handle(event_ref.ref)
                eyear = event.get_date_object().to_calendar(self.cal).get_year()
                place_handle = event.get_place_handle()
                if place_handle:
                    place = dbstate.db.get_place_from_handle(place_handle)
                    if place:
                        longitude = place.get_longitude()
                        latitude = place.get_latitude()
                        latitude, longitude = conv_lat_lon(latitude, longitude, "D.D8")
                        descr = _pd.display(dbstate.db, place)
                        evt = EventType(event.get_type())
                        descr1 = _("%(eventtype)s : %(name)s") % {
                            "eventtype": evt,
                            "name": _nd.display(person),
                        }
                        # place.get_longitude and place.get_latitude return
                        # one string. We have coordinates when the two values
                        # contains non null string.
                        if longitude and latitude:
                            if not self._present_in_places_list(
                                2, str(descr1 + descr + str(evt))
                            ):
                                self._append_to_places_list(
                                    descr,
                                    str(descr1 + descr + str(evt)),
                                    _nd.display(person),
                                    latitude,
                                    longitude,
                                    role,
                                    eyear,
                                    event.get_type(),
                                    person.gramps_id,
                                    place.gramps_id,
                                    event.gramps_id,
                                    fam_id,
                                )
                        else:
                            self._append_to_places_without_coord(place.gramps_id, descr)
            family_list = person.get_family_handle_list()
            for family_hdl in family_list:
                family = self.dbstate.db.get_family_from_handle(family_hdl)
                if family is not None:
                    for event_ref in family.get_event_ref_list():
                        if event_ref:
                            event = dbstate.db.get_event_from_handle(event_ref.ref)
                            role = event_ref.get_role()
                            if event.get_place_handle():
                                place_handle = event.get_place_handle()
                                if place_handle:
                                    place = dbstate.db.get_place_from_handle(
                                        place_handle
                                    )
                                    if place:
                                        longitude = place.get_longitude()
                                        latitude = place.get_latitude()
                                        (latitude, longitude) = conv_lat_lon(
                                            latitude, longitude, "D.D8"
                                        )
                                        descr = _pd.display(dbstate.db, place)
                                        evt = EventType(event.get_type())
                                        (
                                            father_name,
                                            mother_name,
                                        ) = self._get_father_and_mother_name(event)
                                        descr1 = "%s : %s - " % (evt, father_name)
                                        descr1 = "%s%s" % (descr1, mother_name)
                                        eyear = (
                                            event.get_date_object()
                                            .to_calendar(self.cal)
                                            .get_year()
                                        )
                                        if longitude and latitude:
                                            if not self._present_in_places_list(
                                                2, str(descr1 + descr + str(evt))
                                            ):
                                                self._append_to_places_list(
                                                    descr,
                                                    str(descr1 + descr + str(evt)),
                                                    _nd.display(person),
                                                    latitude,
                                                    longitude,
                                                    role,
                                                    eyear,
                                                    event.get_type(),
                                                    person.gramps_id,
                                                    place.gramps_id,
                                                    event.gramps_id,
                                                    family.gramps_id,
                                                )
                                        else:
                                            self._append_to_places_without_coord(
                                                place.gramps_id, descr
                                            )

    def family_label(self, family):
        """
        Create the family label depending on existence of the father and mother
        """
        if family is None:
            return "Unknown"
        father = mother = None
        hdl = family.get_father_handle()
        if hdl:
            father = self.dbstate.db.get_person_from_handle(hdl)
        hdl = family.get_mother_handle()
        if hdl:
            mother = self.dbstate.db.get_person_from_handle(hdl)
        if father and mother:
            label = _("%(gramps_id)s : %(father)s and %(mother)s") % {
                "father": _nd.display(father),
                "mother": _nd.display(mother),
                "gramps_id": family.gramps_id,
            }
        elif father:
            label = "%(gramps_id)s : %(father)s" % {
                "father": _nd.display(father),
                "gramps_id": family.gramps_id,
            }
        elif mother:
            label = "%(gramps_id)s : %(mother)s" % {
                "mother": _nd.display(mother),
                "gramps_id": family.gramps_id,
            }
        else:
            # No translation for bare gramps_id
            label = "%(gramps_id)s :" % {
                "gramps_id": family.gramps_id,
            }
        return label

    def _createmap_for_one_family(self, family):
        """
        Create all markers for one family : all event's places with a lat/lon.
        """
        dbstate = self.dbstate
        self.message_layer.add_message(
            _("Family places for %s") % self.family_label(family)
        )
        person = None
        if family:
            handle = family.get_father_handle()
            if handle:
                person = dbstate.db.get_person_from_handle(handle)
        else:
            return
        family_id = family.gramps_id
        if person is None:  # family without father ?
            handle = family.get_mother_handle()
            if handle:
                person = dbstate.db.get_person_from_handle(handle)
        if person is None:
            handle = self.uistate.get_active("Person")
            if handle:
                person = dbstate.db.get_person_from_handle(handle)
        if person is not None:
            family_list = person.get_family_handle_list()
            if len(family_list) > 0:
                fhandle = family_list[0]  # first is primary
                fam = dbstate.db.get_family_from_handle(fhandle)
                father = mother = None
                handle = fam.get_father_handle()
                if handle:
                    father = dbstate.db.get_person_from_handle(handle)
                if father:
                    comment = _("Father : %(id)s : %(name)s") % {
                        "id": father.gramps_id,
                        "name": _nd.display(father),
                    }
                    self._createpersonmarkers(dbstate, father, comment, family_id)
                handle = fam.get_mother_handle()
                if handle:
                    mother = dbstate.db.get_person_from_handle(handle)
                if mother:
                    comment = _("Mother : %(id)s : %(name)s") % {
                        "id": mother.gramps_id,
                        "name": _nd.display(mother),
                    }
                    self._createpersonmarkers(dbstate, mother, comment, family_id)
                index = 0
                child_ref_list = fam.get_child_ref_list()
                if child_ref_list:
                    for child_ref in child_ref_list:
                        child = dbstate.db.get_person_from_handle(child_ref.ref)
                        if child:
                            index += 1
                            comment = _("Child : %(id)s - %(index)d " ": %(name)s") % {
                                "id": child.gramps_id,
                                "index": index,
                                "name": _nd.display(child),
                            }
                            self._createpersonmarkers(
                                dbstate, child, comment, family_id
                            )
            else:
                comment = _("Person : %(id)s %(name)s has no family.") % {
                    "id": person.gramps_id,
                    "name": _nd.display(person),
                }
                self._createpersonmarkers(dbstate, person, comment, family_id)

    def _createmap(self, handle):
        """
        Create all markers for each people's event in the database which has
        a lat/lon.
        """
        if self.osm is None:
            return
        if not handle:
            return
        self.place_list = []
        self.place_without_coordinates = []
        self.places_found = []
        self.nbplaces = 0
        self.nbmarkers = 0
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        self.message_layer.clear_messages()
        self.message_layer.set_font_attributes(None, None, None)
        if self.dbstate.db.has_family_handle(handle):
            family = self.dbstate.db.get_family_from_handle(handle)
            self._createmap_for_one_family(family)
        else:
            person = self.dbstate.db.get_person_from_handle(handle)
            if not person:
                return
            family_list = person.get_family_handle_list()
            for family_hdl in family_list:
                family = self.dbstate.db.get_family_from_handle(family_hdl)
                if family is not None:
                    self._createmap_for_one_family(family)
        self.sort = sorted(self.place_list, key=operator.itemgetter(3, 4, 6))
        self._create_markers()

    def add_event_bubble_message(self, event, lat, lon, mark, menu):
        """
        Add an item to the popup menu.
        """
        self.itemoption = Gtk.Menu()
        itemoption = self.itemoption
        itemoption.show()
        menu.set_submenu(itemoption)
        modify = Gtk.MenuItem(label=_("Edit Family"))
        modify.show()
        modify.connect("activate", self.edit_family, event, lat, lon, mark)
        itemoption.append(modify)
        modify = Gtk.MenuItem(label=_("Edit Person"))
        modify.show()
        modify.connect("activate", self.edit_person, event, lat, lon, mark)
        itemoption.append(modify)
        modify = Gtk.MenuItem(label=_("Edit Event"))
        modify.show()
        modify.connect("activate", self.edit_event, event, lat, lon, mark)
        itemoption.append(modify)
        center = Gtk.MenuItem(label=_("Center on this place"))
        center.show()
        center.connect("activate", self.center_here, event, lat, lon, mark)
        itemoption.append(center)

    def bubble_message(self, event, lat, lon, marks):
        """
        Add the popup menu.
        """
        self.menu = Gtk.Menu()
        menu = self.menu
        message = ""
        oldplace = ""
        prevmark = None
        for mark in marks:
            if message != "":
                add_item = Gtk.MenuItem(label=message)
                add_item.show()
                menu.append(add_item)
                self.add_event_bubble_message(event, lat, lon, prevmark, add_item)
            if mark[0] != oldplace:
                message = "%s :" % mark[0]
                self.add_place_bubble_message(
                    event, lat, lon, marks, menu, message, mark
                )
                oldplace = mark[0]
            evt = self.dbstate.db.get_event_from_gramps_id(mark[10])
            # format the date as described in preferences.
            date = displayer.display(evt.get_date_object())
            if date == "":
                date = _("Unknown")
            if mark[5] == EventRoleType.PRIMARY:
                message = "(%s) %s : %s" % (date, mark[7], mark[1])
            elif mark[5] == EventRoleType.FAMILY:
                evt = self.dbstate.db.get_event_from_gramps_id(mark[10])
                (father_name, mother_name) = self._get_father_and_mother_name(evt)
                message = "(%s) %s : %s - %s" % (
                    date,
                    mark[7],
                    father_name,
                    mother_name,
                )
            else:
                evt = self.dbstate.db.get_event_from_gramps_id(mark[10])
                descr = evt.get_description()
                if descr == "":
                    descr = _("No description")
                message = "(%s) %s => %s" % (date, mark[5], descr)
            prevmark = mark
        add_item = Gtk.MenuItem(label=message)
        add_item.show()
        menu.append(add_item)
        self.add_event_bubble_message(event, lat, lon, prevmark, add_item)
        menu.popup_at_pointer(event)
        return 1

    def add_specific_menu(self, menu, event, lat, lon):
        """
        Add specific entry to the navigation menu.
        """
        return

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (("Family Filter",), ())



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2024-2025  Steve Youngs
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# python
#
# -------------------------------------------------------------------------
import pickle

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk, Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.clipboard import obj2icon, obj2target
from gramps.gui.ddtargets import DdTargets
from gramps.gen.lib.datebase import DateBase
from gramps.gui.listmodel import ListModel
from gramps.gen.utils.db import navigation_label
from gramps.gen.plug import Gramplet
from gramps.gui.utils import edit_object
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.datehandler import displayer

_ = glocale.translation.gettext


class Backlinks(Gramplet):
    """
    Displays the back references for an object.
    """

    __obj2target = {
        "Person": DdTargets.PERSON_LINK,
        "Family": DdTargets.FAMILY_LINK,
        "Event": DdTargets.EVENT,
        "Citation": DdTargets.CITATION_LINK,
        "Source": DdTargets.SOURCE_LINK,
        "Place": DdTargets.PLACE_LINK,
        "Repository": DdTargets.REPO_LINK,
        "Media": DdTargets.MEDIAOBJ,
        "Note": DdTargets.NOTE_LINK,
    }

    def __init__(self, classname, gui, nav_group=0):
        self.__classname = classname  # must be set before calling super().__init__
        super().__init__(gui, nav_group)

    def init(self):
        self.date_column = None
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.top = Gtk.TreeView()
        self.top.get_selection().connect("changed", self.selection_changed)
        self.top.connect_after("drag-begin", self.drag_begin)
        self.top.connect("drag_data_get", self.drag_data_get)

        titles = [
            (_("Type"), 1, 100),
            (_("Name"), 2, 100),
            (_("Date"), 4, 200),
            ("date_sort_value", 4, 1),  # hidden column
            ("", 5, 1),  # hidden column for the handle
            ("", 6, 1),  # hidden column for non-localized object type
        ]
        self.model = ListModel(
            self.top,
            titles,
            event_func=self.cb_double_click,
            middle_click=self.cb_middle_click,
            right_click=self.cb_right_click,
        )

        self.date_column = self.top.get_column(2)
        self.date_column.set_visible(False)
        self.date_column.set_sort_column_id(3)
        self.top.get_column(1).set_expand(
            True
        )  # Expand the Name column as far as possible
        self.top.get_column(3).set_visible(
            False
        )  # always hide the date_sort_value column
        # possible size
        return self.top

    def selection_changed(self, selection):
        """
        Called when the selection is changed in the TreeView.
        """
        target = None
        (model, iter_) = selection.get_selected()
        if iter_:
            objclass = model.get_value(iter_, 5)
            target = self.__obj2target.get(objclass)

        if target:
            self.top.drag_source_set(
                Gdk.ModifierType.BUTTON1_MASK, [target.target()], Gdk.DragAction.COPY
            )
        else:
            self.top.drag_source_unset()

    def drag_begin(self, widget, drag_context):
        """
        We want to show the icon during drag instead of the long row entry
        """
        (model, iter_) = self.top.get_selection().get_selected()
        if not iter_:
            return

        objclass = model.get_value(iter_, 5)
        Gtk.drag_set_icon_name(drag_context, obj2icon(objclass.lower()), 0, 0)

    def drag_data_get(self, widget, context, sel_data, info, time):
        # get the selected object, returning if nothing is selected
        (model, iter_) = self.top.get_selection().get_selected()
        if not iter_:
            return

        (objclass, handle) = (model.get_value(iter_, 5), model.get_value(iter_, 4))

        target = obj2target(objclass)
        data = (target.name(), id(self), handle, 0)

        sel_data.set(target, 8, pickle.dumps(data))

    def display_backlinks(self, active_handle):
        """
        Display the back references for an object.
        """
        have_dates = False  # True if any of the objects we are displaying are instances of DateBase
        for classname, handle in self.dbstate.db.find_backlink_handles(active_handle):
            name = navigation_label(self.dbstate.db, classname, handle)[0]
            obj = self.dbstate.db.method("get_%s_from_handle", classname)(handle)
            if isinstance(obj, DateBase):
                o_date = obj.get_date_object()
                date = displayer.display(o_date)
                date_sort_value = "%09d" % o_date.get_sort_value()
                have_dates = True
            else:
                date = date_sort_value = ""
            self.model.add(
                (_(classname), name, date, date_sort_value, handle, classname)
            )

        self.date_column.set_visible(
            have_dates
        )  # only show the Date column if there are some values
        self.set_has_data(self.model.count > 0)

    def get_has_data(self, active_handle):
        """
        Return True if the gramplet has data, else return False.
        """
        if not active_handle:
            return False
        for handle in self.dbstate.db.find_backlink_handles(active_handle):
            return True
        return False

    def cb_double_click(self, treeview):
        """
        Handle double click on treeview.
        """
        (model, iter_) = treeview.get_selection().get_selected()
        if not iter_:
            return

        (objclass, handle) = (model.get_value(iter_, 5), model.get_value(iter_, 4))

        edit_object(self.dbstate, self.uistate, objclass, handle)

    def cb_middle_click(self, treeview, event):
        """
        Handle middle click on view.
        """
        (model, iter_) = treeview.get_selection().get_selected()
        if not iter_:
            return

        (objclass, handle) = (model.get_value(iter_, 5), model.get_value(iter_, 4))
        self.uistate.set_active(handle, objclass)

    def cb_right_click(self, treeview, event):
        """
        Handle right click on view.
        """
        (model, iter_) = self.top.get_selection().get_selected()
        if not iter_:
            return  # nothing selected, so do not display the context menu

        (objclass, handle) = (model.get_value(iter_, 5), model.get_value(iter_, 4))

        self.__store_menu = Gtk.Menu()  # need to keep reference or menu disappears
        menu = self.__store_menu

        item = Gtk.MenuItem.new_with_mnemonic(_("_Edit"))
        item.connect(
            "activate",
            lambda obj: edit_object(self.dbstate, self.uistate, objclass, handle),
        )
        item.show()
        menu.append(item)

        item = Gtk.MenuItem.new_with_mnemonic(_("_Make Active %s") % objclass)
        item.connect("activate", lambda obj: self.uistate.set_active(handle, objclass))
        item.show()
        menu.append(item)

        menu.popup_at_pointer(event)
        return True

    def db_changed(self):
        for item in [
            "person",
            "family",
            "source",
            "citation",
            "event",
            "media",
            "place",
            "repository",
            "note",
        ]:
            self.connect(self.dbstate.db, "%s-delete" % item, self.update)
            self.connect(self.dbstate.db, "%s-add" % item, self.update)
            self.connect(self.dbstate.db, "%s-update" % item, self.update)
        if self.__classname != "Person":
            # the Gramplet super class already connects to changes in
            # active person and calls active_changed
            self.connect_signal(self.__classname, self.update)

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active(self.__classname)
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active(self.__classname)
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class PersonBacklinks(Backlinks):
    """
    Displays the back references for a person.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Person", gui, nav_group)


class EventBacklinks(Backlinks):
    """
    Displays the back references for an event.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Event", gui, nav_group)


class FamilyBacklinks(Backlinks):
    """
    Displays the back references for a family.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Family", gui, nav_group)


class PlaceBacklinks(Backlinks):
    """
    Displays the back references for a place.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Place", gui, nav_group)


class SourceBacklinks(Backlinks):
    """
    Displays the back references for a source.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Source", gui, nav_group)


class CitationBacklinks(Backlinks):
    """
    Displays the back references for a Citation.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Citation", gui, nav_group)


class RepositoryBacklinks(Backlinks):
    """
    Displays the back references for a repository.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Repository", gui, nav_group)


class MediaBacklinks(Backlinks):
    """
    Displays the back references for a media object.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Media", gui, nav_group)


class NoteBacklinks(Backlinks):
    """
    Displays the back references for a note.
    """

    def __init__(self, gui, nav_group=0):
        super().__init__("Note", gui, nav_group)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

try:
    set()
except:
    from sets import Set as set

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import List, Set
from ....lib import Person
from ....db import Database
from ....types import PersonHandle


# -------------------------------------------------------------------------
#
# IsLessThanNthGenerationAncestorOfBookmarked
#
# -------------------------------------------------------------------------
class IsLessThanNthGenerationAncestorOfBookmarked(Rule):
    # Submitted by Wayne Bergeron
    """Rule that checks for a person that is an ancestor of bookmarked persons
    not more than N generations away"""

    labels = [_("Number of generations:")]
    name = _("Ancestors of bookmarked people not more " "than <N> generations away")
    category = _("Ancestral filters")
    description = _(
        "Matches ancestors of the people on the bookmark list "
        "not more than N generations away"
    )

    def prepare(self, db: Database, user):
        self.db = db
        bookmarks: List[str] = db.get_bookmarks().get()
        self.selected_handles: Set[PersonHandle] = set()
        if len(bookmarks) != 0:
            self.bookmarks: Set[PersonHandle] = set(bookmarks)
            for self.bookmarkhandle in self.bookmarks:
                self.init_ancestor_list(self.bookmarkhandle, 1)

    def init_ancestor_list(self, handle: PersonHandle, gen: int):
        if not handle or handle in self.selected_handles:
            # if been here already, skip
            return
        if gen:
            self.selected_handles.add(handle)
            if gen >= int(self.list[0]):
                return

        p = self.db.get_person_from_handle(handle)
        fam_id = p.parent_family_list[0] if len(p.parent_family_list) > 0 else None
        if not fam_id:
            return
        fam = self.db.get_family_from_handle(fam_id)
        if fam:
            f_id = fam.father_handle
            m_id = fam.mother_handle

            if f_id:
                self.init_ancestor_list(f_id, gen + 1)
            if m_id:
                self.init_ancestor_list(m_id, gen + 1)

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles

    def reset(self):
        self.selected_handles.clear()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"

# ------------------------------------------------------------------------
#
# EniroMaps
#
# ------------------------------------------------------------------------

register(
    MAPSERVICE,
    id="EniroMaps",
    name=_("EniroMaps"),
    description=_("Opens on kartor.eniro.se"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="eniroswedenmap.py",
    authors=["Peter Landgren"],
    authors_email=["peter.talken@telia.com"],
    mapservice="EniroSVMapService",
)

# ------------------------------------------------------------------------
#
# GoogleMaps
#
# ------------------------------------------------------------------------

register(
    MAPSERVICE,
    id="GoogleMaps",
    name=_("GoogleMaps"),
    description=_("Open on maps.google.com"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="googlemap.py",
    authors=["Benny Malengier"],
    authors_email=["benny.malengier@gramps-project.org"],
    mapservice="GoogleMapService",
)

# ------------------------------------------------------------------------
#
# OpenStreetMap
#
# ------------------------------------------------------------------------

register(
    MAPSERVICE,
    id="OpenStreetMap",
    name=_("OpenStreetMap"),
    description=_("Open on openstreetmap.org"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="openstreetmap.py",
    authors=["Benny Malengier"],
    authors_email=["benny.malengier@gramps-project.org"],
    mapservice="OpensStreetMapService",
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = ["IconButton", "WarnButton", "SimpleButton", "PrivacyButton"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

import logging

_LOG = logging.getLogger(".widgets.buttons")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk


# -------------------------------------------------------------------------
#
# IconButton class
#
# -------------------------------------------------------------------------
class IconButton(Gtk.Button):
    def __init__(self, func, handle, icon="gtk-edit", size=Gtk.IconSize.MENU):
        Gtk.Button.__init__(self)
        image = Gtk.Image()
        image.set_from_icon_name(icon, size)
        image.show()
        self.add(image)
        self.set_relief(Gtk.ReliefStyle.NONE)
        self.show()

        if func:
            self.connect("button-press-event", func, handle)
            self.connect("key-press-event", func, handle)


##    def destroy(self):
##        """
##        Unset all elements that can prevent garbage collection
##        """
##        Gtk.Button.destroy(self)


# -------------------------------------------------------------------------
#
# WarnButton class
#
# -------------------------------------------------------------------------
class WarnButton(Gtk.Button):
    def __init__(self):
        Gtk.Button.__init__(self)

        image = Gtk.Image()
        image.set_from_icon_name("dialog-information", Gtk.IconSize.MENU)
        image.show()
        self.add(image)

        self.set_relief(Gtk.ReliefStyle.NONE)
        self.show()
        self.func = None
        self.hide()

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.func = None
    ##        Gtk.Button.destroy(self)

    def on_clicked(self, func):
        self.connect("button-press-event", self._button_press)
        self.func = func

    def _button_press(self, obj, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 1:
            self.func(obj)


# -------------------------------------------------------------------------
#
# SimpleButton class
#
# -------------------------------------------------------------------------
class SimpleButton(Gtk.Button):
    def __init__(self, image, func):
        Gtk.Button.__init__(self)
        self.set_relief(Gtk.ReliefStyle.NONE)
        self.add(Gtk.Image.new_from_icon_name(image, Gtk.IconSize.BUTTON))
        self.connect("clicked", func)
        self.show()


##    def destroy(self):
##        """
##        Unset all elements that can prevent garbage collection
##        """
##        Gtk.Button.destroy(self)


# -------------------------------------------------------------------------
#
# PrivacyButton class
#
# -------------------------------------------------------------------------
class PrivacyButton:
    def __init__(self, button, obj, readonly=False):
        self.button = button
        self.button.connect("toggled", self._on_toggle)
        self.obj = obj
        self.set_active(obj.get_privacy())
        self.button.set_sensitive(not readonly)

    ##    def destroy(self):
    ##        """
    ##        Unset all elements that can prevent garbage collection
    ##        """
    ##        self.obj = None

    def set_sensitive(self, val):
        self.button.set_sensitive(val)

    def set_active(self, val):
        self.button.set_active(val)
        self._on_toggle(self.button)

    def get_active(self):
        return self.button.get_active()

    def _on_toggle(self, obj):
        child = obj.get_child()
        if child:
            obj.remove(child)
        image = Gtk.Image()
        if obj.get_active():
            image.set_from_icon_name("gramps-lock", Gtk.IconSize.MENU)
            obj.set_tooltip_text(_("Record is private"))
            self.obj.set_privacy(True)
        else:
            image.set_from_icon_name("gramps-unlock", Gtk.IconSize.MENU)
            obj.set_tooltip_text(_("Record is public"))
            self.obj.set_privacy(False)
        image.show()
        obj.add(image)



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016  Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Geography for events
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import operator

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import EventType
from gramps.gen.config import config
from gramps.gen.datehandler import displayer
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd
from gramps.gen.utils.place import conv_lat_lon
from gramps.gui.views.bookmarks import EventBookmarks
from gramps.plugins.lib.maps.geography import GeoGraphyView
from gramps.gui.utils import ProgressMeter

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
KEY_TAB = Gdk.KEY_Tab
_LOG = logging.getLogger("GeoGraphy.geoevents")
_ = glocale.translation.gettext

_UI_DEF = [
    """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
    """,
    """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.PrintView</attribute>
          <attribute name="label" translatable="yes">Print...</attribute>
        </item>
      </section>
    """,
    """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
    """
    % _("Organize Bookmarks"),  # Following are the Toolbar items
    """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text"
         translatable="yes">Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text"
         translatable="yes">Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
    """,
    """
    <placeholder id='BarCommonEdit'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">document-print</property>
        <property name="action-name">win.PrintView</property>
        <property name="tooltip_text"
         translatable="yes">Print or save the Map</property>
        <property name="label" translatable="yes">Print...</property>
        <property name="use-underline">True</property>
       </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
    """,
]

# pylint: disable=unused-argument
# pylint: disable=no-member
# pylint: disable=maybe-no-member


# -------------------------------------------------------------------------
#
# GeoView
#
# -------------------------------------------------------------------------
class GeoEvents(GeoGraphyView):
    """
    The view used to render events map.
    """

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        self.window_name = _("Events places map")
        GeoGraphyView.__init__(
            self, self.window_name, pdata, dbstate, uistate, EventBookmarks, nav_group
        )
        self.dbstate = dbstate
        self.uistate = uistate
        self.place_list = []
        self.place_without_coordinates = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        self.nbplaces = 0
        self.nbmarkers = 0
        self.sort = []
        self.generic_filter = None
        self.additional_uis.append(self.additional_ui())
        self.no_show_places_in_status_bar = False
        self.show_all = False
        self.cal = None

    def get_title(self):
        """
        Used to set the titlebar in the configuration window.
        """
        return _("GeoEvents")

    def get_stock(self):
        """
        Returns the name of the stock icon to use for the display.
        This assumes that this icon has already been registered
        as a stock icon.
        """
        return "geo-show-events"

    def get_viewtype_stock(self):
        """Type of view in category"""
        return "geo-show-events"

    def additional_ui(self):
        """
        Specifies the UIManager XML code that defines the menus and buttons
        associated with the interface.
        """
        return _UI_DEF

    def navigation_type(self):
        """
        Indicates the navigation type. Navigation type can be the string
        name of any of the primary objects.
        """
        return "Event"

    def goto_handle(self, handle=None):
        """
        Rebuild the tree with the given events handle as the root.
        """
        if self.osm is None:
            return
        self.places_found = []
        self.build_tree()

    def show_all_events(self, menu, event, lat, lon):
        """
        Ask to show all events.
        """
        self.show_all = True
        self._createmap(None)

    def build_tree(self):
        """
        This is called by the parent class when the view becomes visible. Since
        all handling of visibility is now in rebuild_trees, see that for more
        information.
        """
        if self.osm is None:
            return
        if not self.dbstate.is_open():
            return
        active = self.uistate.get_active("Event")
        if active:
            self._createmap(active)
        else:
            self._createmap(None)

    def _createmap_for_one_event(self, event):
        """
        Create all markers for each people's event in the database which has
        a lat/lon.
        """
        dbstate = self.dbstate
        if self.nbplaces >= self._config.get("geography.max_places"):
            return
        descr1 = descr2 = ""
        if event:
            place_handle = event.get_place_handle()
            eventyear = event.get_date_object().to_calendar(self.cal).get_year()
        else:
            place_handle = None
        if place_handle:
            place = dbstate.db.get_place_from_handle(place_handle)
            if place:
                descr1 = _pd.display(dbstate.db, place)
                longitude = place.get_longitude()
                latitude = place.get_latitude()
                latitude, longitude = conv_lat_lon(latitude, longitude, "D.D8")
                # place.get_longitude and place.get_latitude return
                # one string. We have coordinates when the two values
                # contains non null string.
                if longitude and latitude:
                    person_list = [
                        dbstate.db.get_person_from_handle(ref_handle)
                        for (ref_type, ref_handle) in dbstate.db.find_backlink_handles(
                            event.handle
                        )
                        if ref_type == "Person"
                    ]
                    if person_list:
                        for person in person_list:
                            if descr2 == "":
                                descr2 = ("%s") % _nd.display(person)
                            else:
                                descr2 = ("%s - %s") % (descr2, _nd.display(person))
                    else:
                        # family list ?
                        family_list = [
                            dbstate.db.get_family_from_handle(ref_handle)
                            for (
                                ref_type,
                                ref_handle,
                            ) in dbstate.db.find_backlink_handles(event.handle)
                            if ref_type == "Family"
                        ]
                        if family_list:
                            for family in family_list:
                                father = mother = None
                                hdle = family.get_father_handle()
                                if hdle:
                                    father = dbstate.db.get_person_from_handle(hdle)
                                hdle = family.get_mother_handle()
                                if hdle:
                                    mother = dbstate.db.get_person_from_handle(hdle)
                                descr2 = ("%(father)s - %(mother)s") % {
                                    "father": (
                                        _nd.display(father)
                                        if father is not None
                                        else "?"
                                    ),
                                    "mother": (
                                        _nd.display(mother)
                                        if mother is not None
                                        else "?"
                                    ),
                                }
                        else:
                            descr2 = _("incomplete or unreferenced event ?")
                    self._append_to_places_list(
                        descr1,
                        None,
                        None,
                        latitude,
                        longitude,
                        descr2,
                        eventyear,
                        event.get_type(),
                        None,  # person.gramps_id
                        place.gramps_id,
                        event.gramps_id,
                        None,
                    )

    def _createmap(self, obj):
        """
        Create all markers for each people's event in the database which has
        a lat/lon.
        """
        if self.osm is None:
            return
        dbstate = self.dbstate
        self.place_list = []
        self.places_found = []
        self.place_without_coordinates = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        self.nbmarkers = 0
        self.nbplaces = 0
        self.without = 0
        self.cal = config.get("preferences.calendar-format-report")
        self.message_layer.clear_messages()
        self.message_layer.clear_font_attributes()
        self.no_show_places_in_status_bar = False
        if self.show_all:
            self.show_all = False
            events_handle = dbstate.db.get_event_handles()
            progress = ProgressMeter(
                self.window_name, can_cancel=False, parent=self.uistate.window
            )
            length = len(events_handle)
            progress.set_pass(_("Selecting all events"), length)
            for event_hdl in events_handle:
                event = dbstate.db.get_event_from_handle(event_hdl)
                self._createmap_for_one_event(event)
                progress.step()
            progress.close()
        elif self.generic_filter:
            user = self.uistate.viewmanager.user
            events_list = self.generic_filter.apply(dbstate.db, user=user)
            progress = ProgressMeter(
                self.window_name, can_cancel=False, parent=self.uistate.window
            )
            length = len(events_list)
            progress.set_pass(_("Selecting all events"), length)
            for event_handle in events_list:
                event = dbstate.db.get_event_from_handle(event_handle)
                self._createmap_for_one_event(event)
                progress.step()
            progress.close()
        else:
            if obj:
                event = dbstate.db.get_event_from_handle(obj)
                self._createmap_for_one_event(event)
            self.message_layer.add_message(
                _(
                    "Right click on the map and select 'show all events'"
                    " to show all known events with coordinates. "
                    "You can use the history to navigate on the map. "
                    "You can use filtering."
                )
            )
        self.sort = sorted(self.place_list, key=operator.itemgetter(3, 4, 6))
        if self.nbmarkers > 500:  # performance issue. Is it the good value ?
            self.no_show_places_in_status_bar = True
        self._create_markers()

    def bubble_message(self, event, lat, lon, marks):
        self.menu = Gtk.Menu()
        menu = self.menu
        message = ""
        oldplace = ""
        prevmark = None
        for mark in marks:
            if message != "":
                add_item = Gtk.MenuItem(label=message)
                add_item.show()
                menu.append(add_item)
                self.itemoption = Gtk.Menu()
                itemoption = self.itemoption
                itemoption.show()
                add_item.set_submenu(itemoption)
                modify = Gtk.MenuItem(label=_("Edit Event"))
                modify.show()
                modify.connect("activate", self.edit_event, event, lat, lon, prevmark)
                itemoption.append(modify)
                center = Gtk.MenuItem(label=_("Center on this place"))
                center.show()
                center.connect("activate", self.center_here, event, lat, lon, prevmark)
                itemoption.append(center)
                evt = self.dbstate.db.get_event_from_gramps_id(mark[10])
                hdle = evt.get_handle()
                bookm = Gtk.MenuItem(label=_("Bookmark this event"))
                bookm.show()
                bookm.connect("activate", self.add_bookmark_from_popup, hdle)
                itemoption.append(bookm)
            if mark[0] != oldplace:
                message = "%s :" % mark[0]
                self.add_place_bubble_message(
                    event, lat, lon, marks, menu, message, mark
                )
                oldplace = mark[0]
            evt = self.dbstate.db.get_event_from_gramps_id(mark[10])
            # format the date as described in preferences.
            date = displayer.display(evt.get_date_object())
            message = "(%s) %s : %s" % (date, EventType(mark[7]), mark[5])
            prevmark = mark
        add_item = Gtk.MenuItem(label=message)
        add_item.show()
        menu.append(add_item)
        self.itemoption = Gtk.Menu()
        itemoption = self.itemoption
        itemoption.show()
        add_item.set_submenu(itemoption)
        modify = Gtk.MenuItem(label=_("Edit Event"))
        modify.show()
        modify.connect("activate", self.edit_event, event, lat, lon, prevmark)
        itemoption.append(modify)
        center = Gtk.MenuItem(label=_("Center on this place"))
        center.show()
        center.connect("activate", self.center_here, event, lat, lon, prevmark)
        itemoption.append(center)
        evt = self.dbstate.db.get_event_from_gramps_id(mark[10])
        hdle = evt.get_handle()
        bookm = Gtk.MenuItem(label=_("Bookmark this event"))
        bookm.show()
        bookm.connect("activate", self.add_bookmark_from_popup, hdle)
        itemoption.append(bookm)
        menu.popup_at_pointer(event)
        return 1

    def add_specific_menu(self, menu, event, lat, lon):
        """
        Add specific entry to the navigation menu.
        """
        menu.append(Gtk.SeparatorMenuItem())
        add_item = Gtk.MenuItem(label=_("Show all events"))
        add_item.connect("activate", self.show_all_events, event, lat, lon)
        menu.append(add_item)
        add_item = Gtk.MenuItem(label=_("Centering on Place"))
        menu.append(add_item)
        self.itemoption = Gtk.Menu()
        itemoption = self.itemoption
        add_item.set_submenu(itemoption)
        oldplace = ""
        for mark in self.sort:
            if mark[0] != oldplace:
                oldplace = mark[0]
                modify = Gtk.MenuItem(label=mark[0])
                modify.connect(
                    "activate", self.goto_place, float(mark[3]), float(mark[4])
                )
                itemoption.append(modify)

    def goto_place(self, obj, lat, lon):
        """
        Center the map on latitude, longitude.
        """
        self.set_center(None, None, lat, lon)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (("Event Filter",), ())



# encoding:utf-8
#
# Gramps - a GTK+/GNOME based genealogy program - Records plugin
#
# Copyright (C) 2008-2011 Reinhard Müller
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2013-2016 Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Standard Python modules
#
# ------------------------------------------------------------------------
import datetime

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.lib import (
    ChildRefType,
    Date,
    Span,
    Name,
    StyledText,
    StyledTextTag,
    StyledTextTagType,
)
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.utils.alive import probably_alive
from gramps.gen.proxy import LivingProxyDb

# ------------------------------------------------------------------------
#
# List of records
#
# ------------------------------------------------------------------------


def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh

RECORDS = [
    (_T_("Youngest living person"), "person_youngestliving", True),
    (_T_("Oldest living person"), "person_oldestliving", True),
    (_T_("Person died at youngest age"), "person_youngestdied", False),
    (_T_("Person died at oldest age"), "person_oldestdied", True),
    (_T_("Person married at youngest age"), "person_youngestmarried", True),
    (_T_("Person married at oldest age"), "person_oldestmarried", True),
    (_T_("Person divorced at youngest age"), "person_youngestdivorced", False),
    (_T_("Person divorced at oldest age"), "person_oldestdivorced", False),
    (_T_("Youngest father"), "person_youngestfather", True),
    (_T_("Youngest mother"), "person_youngestmother", True),
    (_T_("Oldest father"), "person_oldestfather", True),
    (_T_("Oldest mother"), "person_oldestmother", True),
    (_T_("Father with most children"), "person_mostkidsfather", False),
    (_T_("Mother with most children"), "person_mostkidsmother", False),
    (_T_("Father with most grandchildren"), "person_mostgrandkidsfather", False),
    (_T_("Mother with most grandchildren"), "person_mostgrandkidsmother", False),
    (_T_("Couple with most children"), "family_mostchildren", True),
    (_T_("Living couple married most recently"), "family_youngestmarried", True),
    (_T_("Living couple married most long ago"), "family_oldestmarried", True),
    (_T_("Shortest past marriage"), "family_shortest", False),
    (_T_("Longest past marriage"), "family_longest", True),
    (_T_("Couple with smallest age difference"), "family_smallestagediff", True),
    (_T_("Couple with biggest age difference"), "family_biggestagediff", True),
]


# ------------------------------------------------------------------------
#
# Global functions
#
# ------------------------------------------------------------------------
def _good_date(date):
    return date is not None and date.is_valid()


def _find_death_date(db, person):
    death_ref = person.get_death_ref()
    if death_ref:
        death = db.get_event_from_handle(death_ref.ref)
        return death.get_date_object()
    else:
        event_list = person.get_primary_event_ref_list()
        for event_ref in event_list:
            event = db.get_event_from_handle(event_ref.ref)
            if event.get_type().is_death_fallback():
                return event.get_date_object()
    return None


def find_records(
    db,
    filter,
    top_size,
    callname,
    trans_text=glocale.translation.sgettext,
    name_format=None,
    living_mode=LivingProxyDb.MODE_INCLUDE_ALL,
    user=None,
):
    """
    @param trans_text: allow deferred translation of strings
    @type trans_text: a GrampsLocale sgettext instance
    trans_text is a defined keyword (see po/update_po.py, po/genpot.sh)
    :param name_format: optional format to control display of person's name
    :type name_format: None or int
    :param living_mode: enable optional control of living people's records
    :type living_mode: int
    """

    def get_unfiltered_person_from_handle(person_handle):
        if living_mode == LivingProxyDb.MODE_INCLUDE_ALL:
            return db.get_person_from_handle(person_handle)
        else:  # we are in the proxy so get the person before proxy changes
            return db.get_unfiltered_person(person_handle)

    today = datetime.date.today()
    today_date = Date(today.year, today.month, today.day)

    # Person records
    person_youngestliving = []
    person_oldestliving = []
    person_youngestdied = []
    person_oldestdied = []
    person_youngestmarried = []
    person_oldestmarried = []
    person_youngestdivorced = []
    person_oldestdivorced = []
    person_youngestfather = []
    person_youngestmother = []
    person_oldestfather = []
    person_oldestmother = []
    person_mostkidsfather = []
    person_mostkidsmother = []
    person_mostgrandkidsfather = []
    person_mostgrandkidsmother = []

    person_handle_list = db.iter_person_handles()

    # the next "if" will turn person_handle_list from a generator into a
    # list, but only when this code is called from a report (which has a
    # filter) and not when called from a gramplet (which has no filter);
    # so the next line drains the generator and turns it into a list
    # always, so the gramplet can use it later, in the second loop
    person_handle_list = list(person_handle_list)

    if filter:
        person_handle_list = filter.apply(db, person_handle_list, user=user)

    for person_handle in person_handle_list:
        person = db.get_person_from_handle(person_handle)
        unfil_person = get_unfiltered_person_from_handle(person_handle)
        if person is None:
            continue

        # FIXME this should check for a "fallback" birth also/instead
        birth_ref = person.get_birth_ref()

        if not birth_ref:
            # No birth event, so we can't calculate any age.
            continue

        birth = db.get_event_from_handle(birth_ref.ref)
        birth_date = birth.get_date_object()

        death_date = _find_death_date(db, person)

        if not _good_date(birth_date):
            # Birth date unknown or incomplete, so we can't calculate any age.
            continue

        name = _get_styled_primary_name(
            person, callname, trans_text=trans_text, name_format=name_format
        )

        if death_date is None:
            if probably_alive(unfil_person, db):
                # Still living, look for age records
                _record(
                    person_youngestliving,
                    person_oldestliving,
                    today_date - birth_date,
                    name,
                    "Person",
                    person_handle,
                    top_size,
                )
        elif _good_date(death_date):
            # Already died, look for age records
            _record(
                person_youngestdied,
                person_oldestdied,
                death_date - birth_date,
                name,
                "Person",
                person_handle,
                top_size,
            )

        for family_handle in person.get_family_handle_list():
            family = db.get_family_from_handle(family_handle)

            marriage_date = None
            divorce_date = None
            for event_ref in family.get_event_ref_list():
                event = db.get_event_from_handle(event_ref.ref)
                if event.get_type().is_marriage() and (
                    event_ref.get_role().is_family()
                    or event_ref.get_role().is_primary()
                ):
                    marriage_date = event.get_date_object()
                elif event.get_type().is_divorce() and (
                    event_ref.get_role().is_family()
                    or event_ref.get_role().is_primary()
                ):
                    divorce_date = event.get_date_object()

            if _good_date(marriage_date):
                _record(
                    person_youngestmarried,
                    person_oldestmarried,
                    marriage_date - birth_date,
                    name,
                    "Person",
                    person_handle,
                    top_size,
                )

            if _good_date(divorce_date):
                _record(
                    person_youngestdivorced,
                    person_oldestdivorced,
                    divorce_date - birth_date,
                    name,
                    "Person",
                    person_handle,
                    top_size,
                )

            for child_ref in family.get_child_ref_list():
                if person.get_gender() == person.MALE:
                    relation = child_ref.get_father_relation()
                elif person.get_gender() == person.FEMALE:
                    relation = child_ref.get_mother_relation()
                else:
                    continue
                if relation != ChildRefType.BIRTH:
                    continue

                child = db.get_person_from_handle(child_ref.ref)

                # FIXME this should check for a "fallback" birth also/instead
                child_birth_ref = child.get_birth_ref()
                if not child_birth_ref:
                    continue

                child_birth = db.get_event_from_handle(child_birth_ref.ref)
                child_birth_date = child_birth.get_date_object()

                if not _good_date(child_birth_date):
                    continue

                if person.get_gender() == person.MALE:
                    _record(
                        person_youngestfather,
                        person_oldestfather,
                        child_birth_date - birth_date,
                        name,
                        "Person",
                        person_handle,
                        top_size,
                    )
                elif person.get_gender() == person.FEMALE:
                    _record(
                        person_youngestmother,
                        person_oldestmother,
                        child_birth_date - birth_date,
                        name,
                        "Person",
                        person_handle,
                        top_size,
                    )

    for person_handle in person_handle_list:
        # this "person loop" doesn't care about a person's birth or death
        person = db.get_person_from_handle(person_handle)
        if person is None:
            continue

        name = _get_styled_primary_name(
            person, callname, trans_text=trans_text, name_format=name_format
        )

        person_child_list = get_birth_children(db, person)
        if person.get_gender() == person.MALE:
            _record(
                None,
                person_mostkidsfather,
                len(person_child_list),
                name,
                "Person",
                person_handle,
                top_size,
            )
        elif person.get_gender() == person.FEMALE:
            _record(
                None,
                person_mostkidsmother,
                len(person_child_list),
                name,
                "Person",
                person_handle,
                top_size,
            )

        person_grandchild_list = []
        for child in person_child_list:
            person_grandchild_list += get_birth_children(db, child)
        if person.get_gender() == person.MALE:
            _record(
                None,
                person_mostgrandkidsfather,
                len(person_grandchild_list),
                name,
                "Person",
                person_handle,
                top_size,
            )
        elif person.get_gender() == person.FEMALE:
            _record(
                None,
                person_mostgrandkidsmother,
                len(person_grandchild_list),
                name,
                "Person",
                person_handle,
                top_size,
            )

    # Family records
    family_mostchildren = []
    family_youngestmarried = []
    family_oldestmarried = []
    family_shortest = []
    family_longest = []
    family_smallestagediff = []
    family_biggestagediff = []

    for family in db.iter_families():
        # family = db.get_family_from_handle(family_handle)
        if living_mode != LivingProxyDb.MODE_INCLUDE_ALL:
            # FIXME no iter_families method in LivingProxyDb so do it this way
            family = db.get_family_from_handle(family.get_handle())

        father_handle = family.get_father_handle()
        if not father_handle:
            continue
        mother_handle = family.get_mother_handle()
        if not mother_handle:
            continue

        # Test if either father or mother are in filter
        if filter:
            # we don't want many progress reports popping up, so no user=user
            if not filter.apply(db, [father_handle, mother_handle]):
                continue

        father = db.get_person_from_handle(father_handle)
        unfil_father = get_unfiltered_person_from_handle(father_handle)
        if father is None:
            continue
        mother = db.get_person_from_handle(mother_handle)
        unfil_mother = get_unfiltered_person_from_handle(mother_handle)
        if mother is None:
            continue

        father_name = _get_styled_primary_name(
            father, callname, trans_text=trans_text, name_format=name_format
        )
        mother_name = _get_styled_primary_name(
            mother, callname, trans_text=trans_text, name_format=name_format
        )

        name = StyledText(trans_text("%(father)s and %(mother)s"))
        name = name.replace("%(father)s", father_name)
        name = name.replace("%(mother)s", mother_name)

        if living_mode == LivingProxyDb.MODE_INCLUDE_ALL or (
            not probably_alive(unfil_father, db)
            and not probably_alive(unfil_mother, db)
        ):
            _record(
                None,
                family_mostchildren,
                len(family.get_child_ref_list()),
                name,
                "Family",
                family.handle,
                top_size,
            )

        father_birth_ref = father.get_birth_ref()
        if father_birth_ref:
            father_birth_date = db.get_event_from_handle(
                father_birth_ref.ref
            ).get_date_object()
        else:
            father_birth_date = None

        mother_birth_ref = mother.get_birth_ref()
        if mother_birth_ref:
            mother_birth_date = db.get_event_from_handle(
                mother_birth_ref.ref
            ).get_date_object()
        else:
            mother_birth_date = None

        if _good_date(father_birth_date) and _good_date(mother_birth_date):
            if father_birth_date >> mother_birth_date:
                _record(
                    family_smallestagediff,
                    family_biggestagediff,
                    father_birth_date - mother_birth_date,
                    name,
                    "Family",
                    family.handle,
                    top_size,
                )
            elif mother_birth_date >> father_birth_date:
                _record(
                    family_smallestagediff,
                    family_biggestagediff,
                    mother_birth_date - father_birth_date,
                    name,
                    "Family",
                    family.handle,
                    top_size,
                )

        marriage_date = None
        divorce = None
        divorce_date = None
        for event_ref in family.get_event_ref_list():
            event = db.get_event_from_handle(event_ref.ref)
            if event.get_type().is_marriage() and (
                event_ref.get_role().is_family() or event_ref.get_role().is_primary()
            ):
                marriage_date = event.get_date_object()
            if (
                event
                and event.get_type().is_divorce()
                and (
                    event_ref.get_role().is_family()
                    or event_ref.get_role().is_primary()
                )
            ):
                divorce = event
                divorce_date = event.get_date_object()

        father_death_date = _find_death_date(db, father)
        mother_death_date = _find_death_date(db, mother)

        if not _good_date(marriage_date):
            # Not married or marriage date unknown
            continue

        if divorce is not None and not _good_date(divorce_date):
            # Divorced but date unknown or inexact
            continue

        if not probably_alive(unfil_father, db) and not _good_date(father_death_date):
            # Father died but death date unknown or inexact
            continue

        if not probably_alive(unfil_mother, db) and not _good_date(mother_death_date):
            # Mother died but death date unknown or inexact
            continue

        if (
            divorce_date is None
            and father_death_date is None
            and mother_death_date is None
        ):
            # Still married and alive
            if probably_alive(unfil_father, db) and probably_alive(unfil_mother, db):
                _record(
                    family_youngestmarried,
                    family_oldestmarried,
                    today_date - marriage_date,
                    name,
                    "Family",
                    family.handle,
                    top_size,
                )
        elif (
            _good_date(divorce_date)
            or _good_date(father_death_date)
            or _good_date(mother_death_date)
        ):
            end = None
            if _good_date(father_death_date) and _good_date(mother_death_date):
                end = min(father_death_date, mother_death_date)
            elif _good_date(father_death_date):
                end = father_death_date
            elif _good_date(mother_death_date):
                end = mother_death_date
            if _good_date(divorce_date):
                if end:
                    end = min(end, divorce_date)
                else:
                    end = divorce_date
            duration = end - marriage_date

            _record(
                family_shortest,
                family_longest,
                duration,
                name,
                "Family",
                family.handle,
                top_size,
            )
    # python 3 workaround: assign locals to tmp so we work with runtime version
    tmp = locals()
    return [
        (trans_text(text), varname, tmp[varname])
        for (text, varname, default) in RECORDS
    ]


def _record(lowest, highest, value, text, handle_type, handle, top_size):
    if value < 0:  # ignore erroneous data
        return  # (since the data-verification tool already finds it)

    if isinstance(value, Span):
        low_value = value.minmax[0]
        high_value = value.minmax[1]
    else:
        low_value = value
        high_value = value

    if lowest is not None:
        lowest.append((high_value, value, text, handle_type, handle))
        lowest.sort(key=lambda a: a[0])  # FIXME: Ist das lambda notwendig?
        for i in range(top_size, len(lowest)):
            if lowest[i - 1][0] < lowest[i][0]:
                del lowest[i:]
                break

    if highest is not None:
        highest.append((low_value, value, text, handle_type, handle))
        highest.sort(reverse=True)
        for i in range(top_size, len(highest)):
            if highest[i - 1][0] > highest[i][0]:
                del highest[i:]
                break


def get_birth_children(db, person):
    """return all the birth children of a person, in a list"""
    person_child_list = []
    for family_handle in person.get_family_handle_list():
        family = db.get_family_from_handle(family_handle)
        for child_ref in family.get_child_ref_list():
            if person.get_gender() == person.MALE:
                relation = child_ref.get_father_relation()
            elif person.get_gender() == person.FEMALE:
                relation = child_ref.get_mother_relation()
            else:
                continue  # no records are kept for unknown-sex parents
            if relation != ChildRefType.BIRTH:
                continue  # only count birth children
            child = db.get_person_from_handle(child_ref.ref)
            if child not in person_child_list:
                person_child_list += [child]
    return person_child_list


# ------------------------------------------------------------------------
#
# Reusable functions (could be methods of gen.lib.*)
#
# ------------------------------------------------------------------------

CALLNAME_DONTUSE = 0
CALLNAME_REPLACE = 1
CALLNAME_UNDERLINE_ADD = 2


def _get_styled(
    name,
    callname,
    placeholder=False,
    trans_text=glocale.translation.sgettext,
    name_format=None,
):
    """
    Return a StyledText object with the name formatted according to the
    parameters:

    @param callname: whether the callname should be used instead of the first
        name (CALLNAME_REPLACE), underlined within the first name
        (CALLNAME_UNDERLINE_ADD) or not used at all (CALLNAME_DONTUSE).
    @param placeholder: whether a series of underscores should be inserted as a
        placeholder if first name or surname are missing.
    @param trans_text: allow deferred translation of strings
    @type trans_text: a GrampsLocale sgettext instance
    trans_text is a defined keyword (see po/update_po.py, po/genpot.sh)
    :param name_format: optional format to control display of person's name
    :type name_format: None or int
    """

    # Make a copy of the name object so we don't mess around with the real
    # data.
    n = Name(source=name)

    # Insert placeholders.
    if placeholder:
        if not n.first_name:
            n.first_name = "____________"
        if not n.surname:
            n.surname = "____________"

    if n.call:
        if callname == CALLNAME_REPLACE:
            # Replace first name with call name.
            n.first_name = n.call
        elif callname == CALLNAME_UNDERLINE_ADD:
            if n.call not in n.first_name:
                # Add call name to first name.
                # Translators: used in French+Russian, ignore otherwise
                n.first_name = trans_text('"%(callname)s" (%(firstname)s)') % {
                    "callname": n.call,
                    "firstname": n.first_name,
                }

    real_format = name_displayer.get_default_format()
    if name_format is not None:
        name_displayer.set_default_format(name_format)
    text = name_displayer.display_name(n)
    name_displayer.set_default_format(real_format)
    tags = []

    if n.call:
        if callname == CALLNAME_UNDERLINE_ADD:
            # "name" in next line is on purpose: only underline the call name
            # if it was a part of the *original* first name
            if n.call in name.first_name:
                # Underline call name
                callpos = text.find(n.call)
                tags = [
                    StyledTextTag(
                        StyledTextTagType.UNDERLINE,
                        True,
                        [(callpos, callpos + len(n.call))],
                    )
                ]

    return StyledText(text, tags)


def _get_styled_primary_name(
    person,
    callname,
    placeholder=False,
    trans_text=glocale.translation.sgettext,
    name_format=None,
):
    """
    Return a StyledText object with the person's name formatted according to
    the parameters:

    @param callname: whether the callname should be used instead of the first
        name (CALLNAME_REPLACE), underlined within the first name
        (CALLNAME_UNDERLINE_ADD) or not used at all (CALLNAME_DONTUSE).
    @param placeholder: whether a series of underscores should be inserted as a
        placeholder if first name or surname are missing.
    @param trans_text: allow deferred translation of strings
    @type trans_text: a GrampsLocale sgettext instance
    trans_text is a defined keyword (see po/update_po.py, po/genpot.sh)
    :param name_format: optional format to control display of person's name
    :type name_format: None or int
    """

    return _get_styled(
        person.get_primary_name(),
        callname,
        trans_text=trans_text,
        placeholder=placeholder,
        name_format=name_format,
    )



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013       Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Deeper testing of some DateParser internals.
"""

import unittest

from ...utils.grampslocale import GrampsLocale
from ...lib.date import Date


class DateParserTest(unittest.TestCase):
    def setUp(self):
        from .._dateparser import DateParser

        self.parser = DateParser()
        self.parser_RU = GrampsLocale(lang="ru").date_parser

    def assert_map_key_val(self, m, k, v):
        try:
            self.assertEqual(m[k], v)
        except KeyError:
            self.assertTrue(False, list(m.items()))

    def test_month_to_int_jan_is_1(self):
        self.assert_map_key_val(self.parser.month_to_int, "jan", 1)

    def test_prefix_table_for_RU_built(self):
        self.assertIn("ru_RU", self.parser._langs)

    def test_month_to_int_septem_RU_is_9(self):
        self.assert_map_key_val(self.parser.month_to_int, "сентяб", 9)

    def test_hebrew_to_int_av_is_12(self):
        self.assert_map_key_val(self.parser.hebrew_to_int, "av", 12)
        self.assert_map_key_val(self.parser.hebrew_to_int, "ав", 12)  # RU

    def test_french_to_int_thermidor_is_11(self):
        self.assert_map_key_val(self.parser.french_to_int, "thermidor", 11)
        self.assert_map_key_val(self.parser.french_to_int, "термидор", 11)  # RU

    def test_islamic_to_int_ramadan_is_9(self):
        self.assert_map_key_val(self.parser.islamic_to_int, "ramadan", 9)
        self.assert_map_key_val(self.parser.islamic_to_int, "рамадан", 9)  # RU

    def test_persian_to_int_tir_is_4(self):
        self.assert_map_key_val(self.parser.persian_to_int, "tir", 4)
        self.assert_map_key_val(self.parser.persian_to_int, "тир", 4)  # RU

    def test_calendar_to_int_gregorian(self):
        self.assert_map_key_val(
            self.parser.calendar_to_int, "gregorian", Date.CAL_GREGORIAN
        )
        self.assert_map_key_val(self.parser.calendar_to_int, "g", Date.CAL_GREGORIAN)
        self.assert_map_key_val(
            self.parser.calendar_to_int, "григорианский", Date.CAL_GREGORIAN
        )
        self.assert_map_key_val(self.parser.calendar_to_int, "г", Date.CAL_GREGORIAN)

    def test_calendar_to_int_julian(self):
        self.assert_map_key_val(self.parser.calendar_to_int, "julian", Date.CAL_JULIAN)
        self.assert_map_key_val(self.parser.calendar_to_int, "j", Date.CAL_JULIAN)
        self.assert_map_key_val(
            self.parser.calendar_to_int, "юлианский", Date.CAL_JULIAN
        )
        self.assert_map_key_val(self.parser.calendar_to_int, "ю", Date.CAL_JULIAN)

    def test_quarter_1(self):
        date = self.parser.parse("q1 1900")
        self.assertTrue(date.is_equal(self.parser.parse("Q1 1900")))
        self.assertEqual(date.get_ymd(), (1900, 1, 1))
        self.assertEqual(date.get_stop_ymd(), (1900, 3, 31))
        self.assertEqual(date.get_modifier(), Date.MOD_RANGE)

    def test_quarter_2(self):
        date = self.parser.parse("q2 1900")
        self.assertTrue(date.is_equal(self.parser.parse("Q2 1900")))
        self.assertEqual(date.get_ymd(), (1900, 4, 1))
        self.assertEqual(date.get_stop_ymd(), (1900, 6, 30))
        self.assertEqual(date.get_modifier(), Date.MOD_RANGE)

    def test_quarter_3(self):
        date = self.parser.parse("q3 1900")
        self.assertTrue(date.is_equal(self.parser.parse("Q3 1900")))
        self.assertEqual(date.get_ymd(), (1900, 7, 1))
        self.assertEqual(date.get_stop_ymd(), (1900, 9, 30))
        self.assertEqual(date.get_modifier(), Date.MOD_RANGE)

    def test_quarter_4(self):
        date = self.parser.parse("q4 1900")
        self.assertTrue(date.is_equal(self.parser.parse("Q4 1900")))
        self.assertEqual(date.get_ymd(), (1900, 10, 1))
        self.assertEqual(date.get_stop_ymd(), (1900, 12, 31))
        self.assertEqual(date.get_modifier(), Date.MOD_RANGE)

    def test_quarter_quality_calendar(self):
        date = self.parser.parse("calc q1 1900 (julian)")
        self.assertEqual(date.get_quality(), Date.QUAL_CALCULATED)
        self.assertEqual(date.get_calendar(), Date.CAL_JULIAN)


class Test_generate_variants(unittest.TestCase):
    def setUp(self):
        from .. import _datestrings
        from .._dateparser import _generate_variants

        self.ds = ds = _datestrings.DateStrings(GrampsLocale(languages=("ru")))
        self.month_variants = list(
            _generate_variants(
                zip(ds.long_months, ds.short_months, ds.swedish_SV, ds.alt_long_months)
            )
        )

    def testVariantsSameLengthAsLongMonths(self):
        self.assertEqual(len(self.ds.long_months), len(self.month_variants))

    def testRussianHasDifferentVariantsForEachMonth(self):
        for i in range(1, 13):
            mvi = self.month_variants[i]
            self.assertTrue(len(mvi) > 1, msg=mvi)

    def testNoEmptyStringInVariants(self):
        for i in range(1, 13):
            mvi = self.month_variants[i]
            self.assertNotIn("", mvi)

    def testLongMonthsAppearInVariants(self):
        for i in range(1, 13):
            lmi = self.ds.long_months[i]
            mvi = self.month_variants[i]
            self.assertIn("{}".format(lmi), mvi)

    def testShortMonthsAppearInVariants(self):
        for i in range(1, 13):
            smi = self.ds.short_months[i]
            mvi = self.month_variants[i]
            self.assertIn("{}".format(smi), mvi)

    def testLongMonthVariantsUnique(self):
        for i in range(1, 13):
            mvi = self.month_variants[i]
            self.assertEqual(len(mvi), len(set(mvi)), msg=mvi)

    def testRuMayVariantsContainSvMaj(self):
        v = self.month_variants[5]
        self.assertIn("Maj", v)


if __name__ == "__main__":
    unittest.main()
